/**************************************************************/
/* blanczos.c                                                 */
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "prand.h"


#define BIT(_i) ((0x00000001)<<(_i))

/* I didn't want to include ggnfs.h for this one lousy prototype.
*/
void printTmp(char *fmt, ...);


/*************************************************************/ 
/* Implementation of the Block Lanczos algorithm for finding */
/* vectors in the kernel of a large sparse matrix.           */
/* The implementation here is as descibed in:                */
/* ``A Block Lanczos Algorithm for Finding Dependencies over */
/*   GF(2)'', Peter Montgomery.                              */
/* The paper I have does not have a publication source on it,*/
/* so it's probably a preprint.                              */
/*************************************************************/ 
/* This version uses the sparse matrix as a black box, so    */
/* this code should be useful to anyone who needs such code  */
/* on a 32-bit platform.                                     */
/* In particular, note that we don't even #include "ggnfs.h" */
/* so that it can be easily taken out and re-used.           */
/*************************************************************/ 


typedef void (* MAT_MULT_FUNC_PTR32)(long *Product, long *x, void *P);
int blockLanczos32(long *deps, MAT_MULT_FUNC_PTR32 MultB, 
                   MAT_MULT_FUNC_PTR32 MultB_T, void *P, long n);

/*****************************************************************/ 
/* So here's what to do: Suppose you have a sparse mxn matrix    */
/* 'B' over F_2 with m<n and need some vectors in the kernel.    */
/* Make two functions like these:                                */
/*                                                               */
/*   void multB( long *Product, long *x, void *P)                */
/*   void multBT(long *Product, long *y, void *P)                */
/*                                                               */
/* Which can compute the products:                               */
/* Product <-- Bx,    Product <-- (B^T)y, respectively,          */
/* where 'x' is an nx32 matrix over F_2,                         */
/*       'y' is an mx32 matrix over F_2.                         */
/* The void pointer 'P' is so you can supply a pointer to        */
/* the data defining the matrix as well (i.e., in our case,      */
/* it will point to the relations). It could be NULL if you don't*/
/* need it. Anyway, once you have such a function and need some  */
/* vectors in the kernel, do:                                    */
/*                                                               */
/*  long deps[n];                                                */
/*  myMat_t *P;                                                  */
/*   ...                                                         */
/*  numDeps = blockLanczos(deps, myMultFunction, (void *)P, n);  */
/*                                                               */
/* to get upto 32 dependencies, stored as column vectors in deps.*/
/* Then if numDeps > 0, you will have 'numDeps' vectors in the   */
/* kernel, stored in a bit-packed fashion. For example, the 1st  */
/* one will be:                                                  */
/*  deps[0]&BIT(0), deps[1]&BIT(0),..., deps[n-1]&BIT(0).        */
/*****************************************************************/ 
/* Note: In the case of a matrix from NFS/MPQS, it amounts       */
/* to 'B' being a matrix whose columns are the relations, so that*/
/* we need a dependency among the columns. Although it is        */
/* sometimes tempting to think of the matrix whose rows are the  */
/* relations, it really is much more convenient to do it this    */
/* way.                                                          */
/*****************************************************************/ 
/* It is assumed that 'long' has 32-bits. It might work if there */
/* are more, but certainly not if there are less. If there are   */
/* more, you're probably better off adjusting the code anyway.   */
/*****************************************************************/ 

/***********************************************************************/
/* Finally note: This code definitely will admit further optimization. */
/* There are several things to do:                                     */
/* (1) Look at the main Lanczos loop itself, and see if any of the     */
/*     caluculations can be (algebraically) factored out. I am probably*/
/*     still wasting some intermediate results and such.               */
/* (2) Run it on a moderate case, like 150Kx150K, and do a complete    */
/*     profile (e.g., with `gprof') to see where most time is being    */
/*     spent. Surely it is in the two functions MultB and MultBT, but  */
/*     it doesn't hurt to check. Besides, you can then get some idea   */
/*     of how much optimizing other functions may have an impact. But  */
/*     the two functions MultB and MultBT can surely be further        */
/*     optimized.                                                      */
/* (3) Look into SSE2/MMX coding up MultB and MultBT to work on 128    */
/*     bit words instead of 32. If it can be done with little added    */
/*     expense, this will cut the number of needed Lanczos iterations  */
/*     down to 1/4 of how many are needed at 32 bits. (It will require */
/*     some changes to other functions as well, but the other changes  */
/*     are straightforward).                                           */
/***********************************************************************/

/***** Prototypes for locally used functions. *****/
void multT(long *res, long *A, long *B, long n);
void multS(long *D, int *S);
void mult32x32(long *res, long *A, long *B);
void preMult(long *A, long *B);
void multnx32(long *C_n, long *A_n, long *B, long n);
void getW_S(long *Wi, int *Si, long *T, int *Si_1);
int  isZeroV(long *A, long size);
int  doColumnOps(long *A, long *B, long n);
int  doColumnOps32(long *A, long *B, long n);

void initbtab();
/**************************************************/


void seedBlockLanczos(long seed)
{ prandseed(seed, 712*seed + 21283, seed^0xF3C91D1A);
}

/**********************************************************************/
int blockLanczos32(long *deps, MAT_MULT_FUNC_PTR32 MultB, 
                  MAT_MULT_FUNC_PTR32 MultB_T, void *P, long n)
/**********************************************************************/
{ long *Y=NULL, *X=NULL, *Vi=NULL, *Vi_1=NULL, *Vi_2=NULL, *tmp_n=NULL, *tmp2_n=NULL;
  long *V0=NULL, *tmp3_n=NULL, *Z=NULL, *AZ=NULL;
  long D[32], E[32], F[32], Wi[32], Wi_1[32], Wi_2[32], T[32], T_1[32], tmp[32];
  long U[32], U_1[32], tmp2[32];
  int  Si[32], Si_1[32];
  long i, j, m, mask, isZero, iterations;
  int  errs=0, numDeps=-1, cont, s;

  
  /* Memory allocation: */
  if (!(Y = (long *)malloc(n*sizeof(long))))    errs++;
  if (!(X = (long *)malloc(n*sizeof(long))))    errs++;
  if (!(Vi = (long *)malloc(n*sizeof(long))))   errs++;
  if (!(V0 = (long *)malloc(n*sizeof(long))))   errs++;
  if (!(Vi_1 = (long *)malloc(n*sizeof(long)))) errs++;
  if (!(Vi_2 = (long *)malloc(n*sizeof(long)))) errs++;
  if (!(tmp_n = (long *)malloc(n*sizeof(long)))) errs++;
  if (!(tmp2_n = (long *)malloc(n*sizeof(long)))) errs++;
  if (!(tmp3_n = (long *)malloc(n*sizeof(long)))) errs++;

  if (errs)
    goto SHORT_CIRC_STOP;
  initbtab();
  
  /******************************************************************/
  /* Throughout, 'A' means the matrix A := (B^T)B. In fact, all the */
  /* notation is directly from Montgomery's paper, except that I    */
  /* don't bother indexing the D_i, E_i, F_i. They are simply D,E,F.*/
  /******************************************************************/
  
  /* Initialization: */
  for (j=0; j<n; j++) {
    Y[j] = prand();  /* Any 32-bit PRNG will do. */
    X[j] = Vi_1[j] = Vi_2[j] = tmp_n[j] = tmp2_n[j] = tmp3_n[j] = 0;
  }
  for (i=0; i<32; i++) {
    Wi[i] = Wi_1[i] = Wi_2[i] = 0;
    T[i] = T_1[i] = U[i] = U_1[i] = tmp[i] = 0;
    Si[i] = Si_1[i] = i; /* Si and Si_1 are both I_32. */
  }
  MultB(tmp_n, Y, P); 
  MultB_T(V0, tmp_n, P);
  memcpy(Vi, V0, n*sizeof(long));

  /* Prime 'T' for the loop, so that we always have */
  /* T = (Vi^T)A(Vi) and T_1 = (Vi_1^T)A(Vi_1).     */
  
  MultB(tmp2_n, Vi, P);
  MultB_T(tmp_n, tmp2_n, P);  /* tmp_n <-- A*Vi */
  multT(T, Vi, tmp_n, n);      /* T <-- (Vi^T)(tmp_n) = (Vi^T)A(Vi) */
  
  cont = 1;
  i = 0;
  getW_S(Wi, Si, T, Si_1); /* Compute W0 and S0. */
  /* Initialize X <-- (V0)(W0)(V0^T)(V0). */
  multT(tmp, V0, V0, n);    /* tmp <-- (V0^T)(V0). */
  mult32x32(tmp2, Wi, tmp); /* tmp2 <-- (W0)(tmp) = (W0)(V0^T)(V0). */
  multnx32(X, V0, tmp2, n); /* X <-- V0(tmp2). */
  iterations = 0;
  do {
    /* Iteration step. */
    iterations++;

    /********** Compute D_{i+1}. **********/
    /* tmp_n = A*Vi from initialization, or the previous iteration. */

    multT(U, tmp_n, tmp_n, n);  /* U <-- (tmp_n)^T(tmp_n) = (Vi^T)(A^2)(Vi) */
    multS(U, Si);            /* U <-- (Vi^T)(A^2)(Vi)(Si)(Si^T). */
    memcpy(D, U, 32*sizeof(long)); /* D <-- U. */
    for (j=0; j<32; j++)
      D[j] ^= T[j]; /* D <-- D + (Vi^T)A(Vi). */
    preMult(D, Wi); /* D <-- (Wi)D. */
    for (j=0; j<32; j++)
      D[j] ^= BIT(j); /* D <-- D + I_{32}. */

    /********** Compute E_{i+1}. **********/
    mult32x32(E, Wi_1, T); /* E <-- (Wi_1)(Vi^T)A(Vi). */
    multS(E, Si);          /* E <-- E(Si)(Si^T).       */

    /**************** Compute F_{i+1}. *************/
    /* Note: We should, at this point, always have */
    /* T_1 = (Vi_1^T)A(Vi_1) and                   */
    /* U_1 = (Vi_1^T)A^2(Vi_1)(Si_1)(Si_1^T)       */
    /***********************************************/
    for (j=0; j<32; j++)
      F[j] = U_1[j] ^ T_1[j];
    multS(F, Si);

    mult32x32(tmp, T_1, Wi_1); /* tmp <-- (Vi_1^T)A(Vi_1)(Wi_1). */
    for (j=0; j<32; j++)
      tmp[j] ^= BIT(j); /* tmp <-- tmp + I_32. */
    preMult(tmp, Wi_2);
    preMult(F, tmp);
    /*** Done computing 'F'. */

    /* Finally, do the actual iteration step, putting V_{i+1} in 'tmp_n'. */
    /* Right now, we still have tmp_n = A*Vi, so we should first use it.  */
    
    /* Do tmp_n <-- tmp_n*(Si)(Si^T). */
    mask = 0x00000000;
    for (j=0; j<32; j++) {
      s = Si[j];
      if ((s>=0) && (s<32))
        mask |= BIT(s);
    }
    for (j=0; j<n; j++) 
      tmp_n[j] &= mask;
    
    multnx32(tmp2_n, Vi, D, n);   /* tmp2_n <-- (Vi)D.        */
    for (j=0; j<n; j++)
      tmp_n[j] ^= tmp2_n[j];   /* tmp_n <-- tmp_n + tmp2_n */
    
    multnx32(tmp2_n, Vi_1, E, n); /* tmp2_n <-- (Vi_1)E */
    for (j=0; j<n; j++)
      tmp_n[j] ^= tmp2_n[j];   /* tmp_n <-- tmp_n + tmp2_n */

    multnx32(tmp2_n, Vi_2, F, n); /* tmp2_n <-- (Vi_2)F */
    for (j=0; j<n; j++)
      tmp_n[j] ^= tmp2_n[j];   /* tmp_n <-- tmp_n + tmp2_n */
    /*** Done. We now have tmp_n = V_{i+1}. ***/
   

    i++;
    memcpy(Vi_2, Vi_1, n*sizeof(long));
    memcpy(Vi_1, Vi, n*sizeof(long));
    memcpy(Vi, tmp_n, n*sizeof(long));
    memcpy(Wi_2, Wi_1, 32*sizeof(long));
    memcpy(Wi_1, Wi, 32*sizeof(long));
    memcpy(T_1, T, 32*sizeof(long));
    memcpy(U_1, U, 32*sizeof(long));
    memcpy(Si_1, Si, 32*sizeof(int));
    
    /******** My `step 3'. ********/
    /* Is (Vi^T)(A)(Vi) == 0 ? */
    MultB(tmp2_n, Vi, P);
    MultB_T(tmp_n, tmp2_n, P);  /* tmp_n <-- A*Vi       */

    multT(T, Vi, tmp_n, n); /* T <-- (Vi^T)A(Vi). */
    getW_S(Wi, Si, T, Si_1);
    
    if (!(isZeroV(T, 32))) {
      /* X <-- X + (Vi)(Wi)(Vi^T)(V_0) */
      multT(tmp, Vi, V0, n); /* tmp <-- (Vi^T)(V0). */
      preMult(tmp, Wi);   /* tmp <-- (Wi)(tmp).  */
      
      multnx32(tmp2_n, Vi, tmp, n);   /* tmp2_n <-- (Vi)(tmp)    */
      for (j=0; j<n; j++)
        X[j] ^= tmp2_n[j]; /* X <-- X + tmp2_n. */
    } else {
      cont=0;
      m = i;
    }
    printTmp("Lanczos: Estimate %1.1lf%% complete...",
              (double)100.0*32.0*iterations/n);  
  } while (cont);
  printf("\nBlock Lanczos used %ld iterations.\n", iterations);

          
  Z = (long *)malloc(2*n*sizeof(long));
  AZ = (long *)malloc(2*n*sizeof(long));
  if (!(Z&&AZ)) {
    fprintf(stderr, "blanczos(): Memory allocation error!\n");
    goto SHORT_CIRC_STOP;
  }
  
  if (isZeroV(Vi, n)) {
    /* Then <X+Y> < ker(A). Later, we will expect AX=0, so do X <-- X+Y. */
    printf("After Block Lanczos iteration, Vm=0.\n");

    for (i=0; i<n; i++)
      X[i] ^= Y[i];
  } else {
    printf("After Block Lanczos iteration, Vm is nonzero. Finishing...\n");
    /* We need more memory (for convenience), so free off what we don't need. */
    free(V0); free(Vi_1); free(Vi_2); free(tmp3_n);
    V0 = Vi_1 = Vi_2 = tmp3_n = NULL;

    /* Construct Z=[ X+Y | Vi] and compute AZ=[ A(X+Y) | AVi ] */
    /* X <-- X+Y, for convenience.                     */
    for (j=0; j<n; j++) {
      X[j] ^= Y[j];
      Z[2*j] = X[j];
      Z[2*j+1] = Vi[j];
    }
    /* Now, compute AZ = [ AX | AVi ]. */
    MultB(tmp2_n, X, P);  
    MultB_T(tmp_n, tmp2_n, P); /* tmp_n <-- AX       */
    for (j=0; j<n; j++) 
      AZ[2*j] = tmp_n[j];
    MultB(tmp2_n, Vi, P);  
    MultB_T(tmp_n, tmp2_n, P); /* tmp_n <-- AVi       */
    for (j=0; j<n; j++) 
      AZ[2*j+1] = tmp_n[j];
    /* Now, AZ should have small rank, and we should be  */
    /* able to do column ops to make zero vectors. Doing */
    /* the same ops to 'Z', we get vectors in ker(A).    */
    doColumnOps(Z, AZ, n);
    /* Now, look for zero columns in AZ. If a column is zero, */
    /* copy the corresponding column of Z into 'X'.           */
    for (i=0; i<n; i++)
      X[i] = 0;
    numDeps=0;
    for (i=0; (i<64) && (numDeps < 32); i++) {
      j=0;
      while ((j<n) && ((AZ[2*j + i/32]&BIT(i%32))==0))
        j++;
      if (j==n) {
        /* Copy column i of Z into column 'numDeps' of X. */
        for (j=0; j<n; j++) {
          if (Z[2*j + i/32]&BIT(i%32))
            X[j] ^= BIT(numDeps);
        }
        numDeps++;
      }
    }
    printf("Found %d dependencies for A=(B^T)B.\n", numDeps);
  }
  j=0;
  while ((j<n)&&(X[j]==0))
    j++;
  if (j==n)
    printf("Probable error: The matrix X is identically zero!!!\n");


  printf("Getting dependencies for original matrix, B...\n");
  /***************************************************/
  /* At this point, we should have AX = (B^T)BX = 0. */
  /* Do simultaneous column ops to BX and X, putting */
  /* BX in RREF, to get vectors in ker(B).           */
  /***************************************************/

  for (i=0; i<n; i++)
    deps[i] = 0;
  numDeps=0;

  MultB(tmp_n, X, P);  /* tmp_n <-- BX. */
  doColumnOps32(X, tmp_n, n);
  for (i=0; i<32; i++) {
    for (j=0, isZero=1; j<n; j++)
      if (tmp_n[j]&BIT(i)) {
        isZero=0;
      }
    if (isZero) {
      for (j=0, isZero=1; j<n; j++) {
        if (X[j]&BIT(i)) {
          deps[j] ^= BIT(numDeps);
          isZero=0;
        }
      }
      if (!(isZero))
        numDeps++;
    }
  }

  if (numDeps) {
    printf("Found %d dependencies for 'B'. Verifying...\n", numDeps);
  } else {
    printf("Some error occurred: all dependencies found seem to be trivial!\n");
    printf("Is Rank(A) = Rank((B^T)B) too small?\n");
    goto SHORT_CIRC_STOP;
  }

  MultB(tmp_n, deps, P);
  for (i=0, isZero=1; (i<n)&&(isZero); i++)
    if (tmp_n[i])
      isZero = 0;
  if (!(isZero))
    printf("Some error occurred: Final product (B)(deps) is nonzero (i=%ld)!\n", i);
  else
    printf("Verified.\n"); 



SHORT_CIRC_STOP:
  
  if (Y != NULL)      free(Y);
  if (X != NULL)      free(X);
  if (Vi != NULL)     free(Vi);
  if (V0 != NULL)     free(V0);
  if (Vi_1 != NULL)   free(Vi_1);
  if (Vi_2 != NULL)   free(Vi_2);
  if (tmp_n != NULL)  free(tmp_n);
  if (tmp2_n != NULL) free(tmp2_n);
  if (tmp3_n != NULL) free(tmp3_n);
  if (Z != NULL)      free(Z);
  if (AZ != NULL)     free(AZ);
  return numDeps;
}

/***********************************************/
int btab[8][128];
/***********************************************/
void initbtab()
/***********************************************/
{ static int initialized=0;
  int i, j, b, s;


  if (initialized) return;
  for (j=0; j<8; j++) {
    b = 1<<j;
    for (i=0,s=0; i<256; i++) {
      if (i&b) btab[j][s++]=i;
    }
  }
  initialized=1;
}

long C[4][256];
/*******************************************/
void multT(long *res, long *A, long *B, long n)
/*******************************************/
/* Input: A and B are nx32 matrices.       */
/* Output: res <-- (A^T)B, a 32x32 matrix. */
/*******************************************/
{ long i, k, l, w;

  for (i=0; i<4; i++)
    memset(C[i], 0x00, 256*sizeof(long));
  for (i=0; i<n; i++) {
    C[0][A[i]&0x000000FF] ^= B[i];
    C[1][(A[i]&0x0000FF00)>>8] ^= B[i];
    C[2][(A[i]&0x00FF0000)>>16] ^= B[i];
    C[3][(A[i]&0xFF000000)>>24] ^= B[i];
  }
  for (k=0; k<4; k++) {
    for (l=0; l<8; l++) {
      w=0x00000000;
      for (i=0; i<128; i++)
        w ^= C[k][btab[l][i]];
      res[8*k+l]=w;
    }
  }
}
  
/*********************************************/
void multS(long *D, int *S)
/*********************************************/
/* Input: D is a 32x32 matrix,               */
/*        'S' is a subset of the columns,    */
/*        {0,1,2,...,31}.                    */
/* Output: Columns of 'D' not in 'S' are     */
/*         zeroed, and the others unchanged. */
/*********************************************/
{ int i, s, mask;

  mask = 0x00000000;
  for (i=0; i<32; i++) {
    s = S[i];
    if ((s>=0) && (s<32))
      mask |= BIT(s);
  }
  for (i=0; i<32; i++) 
    D[i] &= mask;
}

/******************************************/
void mult32x32(long *res, long *A, long *B)
/******************************************/
/* Input: 'A' and 'B' are 32x32 matrices. */
/* Output: res <-- A*B.                   */
/******************************************/
{ int i, j;
  long thisA, thisRes;

  for (i=0; i<32; i++) {
    thisA = A[i]; thisRes = 0;
    for (j=0; j<32; j++) {
      if (thisA&BIT(j))
        thisRes ^= B[j];
    }
    res[i] = thisRes;
  }
}

/******************************************/
void preMult(long *A, long *B)
/******************************************/
/* Input: 'A' and 'B' are 32x32 matrices. */
/* Output: A <-- B*A.                     */
/******************************************/
{ long res[32];

  mult32x32(res, B, A);
  memcpy(A, res, 32*sizeof(long));
}
  
/*********************************************/
void multnx32(long *C_n, long *u_n, long *w, long n)
/*********************************************/
/* Input: u_n is an nx32 matrix,             */
/*          w is a 32x32 matrix.             */
/* Output: C_n is the nx32 product (u_n)(w). */
/*********************************************/
{ long i, j, k, u;

  for (i=0; i<4; i++)
    memset(C[i], 0x00, 256*sizeof(long));

  for (i=0; i<8; i++) {
    for (k=0; k<128; k++) {
      j = btab[i][k];
      C[0][j] ^= w[i];
      C[1][j] ^= w[8+i];
      C[2][j] ^= w[16+i];
      C[3][j] ^= w[24+i];
    }
  }
  for (i=0; i<n; i++) {
    u=u_n[i];
    C_n[i] = C[0][u&0x000000FF]^ C[1][(u&0x0000FF00)>>8] ^
             C[2][(u&0x00FF0000)>>16] ^ C[3][(u&0xFF000000)>>24];
  }
}

/*********************************************/ 
int isZeroV(long *A, long size)
/*********************************************/ 
{ long i;

  for (i=0; i<size; i++)
    if (A[i])
      return 0;
  return 1;
}

/*********************************************/ 
void getW_S(long *Wi, int *Si, long *T, int *Si_1)
{ long M[64], mask, t0, t1;
  int  c[32], i, j, k, sSize, s;
  
  /* Number the columns, with those in Si_1 coming last. */  
  for (sSize=0, mask=0x00000000, i=0; i<32; i++) {
    s = Si_1[i];
    if ((s>=0) && (s<32)) {
      mask |= BIT(s);
      sSize++;
    }
  }
  for (i=0, j=0, k=0; i<32; i++) {
    if (mask&BIT(i)) 
      c[32 - sSize + j++] = i;
    else
      c[k++] = i;
  }
  /* S <-- empty set. */
  sSize=0;
  for (i=0; i<32; i++)
    Si[i] = -1;

  /* Set M <-- [ T | I_32] */
  for (i=0; i<32; i++) {
    M[2*i] = T[i];
    M[2*i+1] = BIT(i);
  }
  for (j=0; j<32; j++) {
    for (k=j; (k<32)&& ( (M[2*c[j]]&BIT(c[j]))==0); k++) {
      if (M[2*c[k]]&BIT(c[j])) {
        /* Exchange rows c[j] and c[k] of M */
        t0 = M[2*c[j]];  t1 = M[2*c[j]+1];
        M[2*c[j]] = M[2*c[k]]; M[2*c[j]+1] = M[2*c[k]+1];
        M[2*c[k]] = t0; M[2*c[k]+1] = t1;
      }
    }
    if ((M[2*c[j]]&BIT(c[j]))) {
      Si[sSize++] = c[j];
      /* Now, add row c[j] to other rows as necessary, to zero out */
      /* the rest of column c[j].                                  */
      for (k=0; k<32; k++) {
        if ((k != c[j]) && (M[2*k]&BIT(c[j]))) {
          M[2*k] ^= M[2*c[j]];
          M[2*k+1] ^= M[2*c[j]+1];
        }
      }
    } else {
      for (k=j; (k<32) && ( (M[2*c[j]+1]&BIT(c[j]))==0); k++) {
        if (M[2*c[k]+1]&BIT(c[j])) {
          /* Exchange rows c[j] and c[k] of M */
          t0 = M[2*c[j]];  t1 = M[2*c[j]+1];
          M[2*c[j]] = M[2*c[k]]; M[2*c[j]+1] = M[2*c[k]+1];
          M[2*c[k]] = t0; M[2*c[k]+1] = t1;
        }
      }
      /* Now, add row c[j] to other rows as necessary, to zero out */
      /* the rest of column c[j]+32                                */
      for (k=0; k<32; k++) {
        if ((k != c[j]) && (M[2*k+1]&BIT(c[j]))) {
          M[2*k] ^= M[2*c[j]];
          M[2*k+1] ^= M[2*c[j]+1];
        }
      }
      /* Zero out row c[j]. */
      M[2*c[j]] = 0; M[2*c[j]+1] = 0;
    } /* end 'else'  */
  } /* end 'for j' */
  for (j=0; j<32; j++)
    Wi[j] = M[2*j+1];
}

/*******************************************************/
int doColumnOps(long *A, long *B, long n)
/*******************************************************/
/* Do column operations (Gaussian elimination) on 'B', */
/* applying the same operations to 'A' as we go.       */
/* Input: 'A' and 'B' are nx64 matrices.               */
/* Output: 'A' and 'B' are changed, as above. i.e., 'B'*/
/*         is in column-reduced echelon form, and the  */
/*         same transformation was applied to 'A'.     */
/*******************************************************/
/* Remark: This is horribly inefficient. There is      */
/* surely a better way to do it. But the easiest is to */
/* just transpose the matrices before and after.       */
/* But it doesn't much matter either way - this guy is */
/* not called from inside any loops, and it's not a    */
/* bottleneck.                                         */
/*******************************************************/
{ long *AT, *BT;
  long i, j, k, rowSize, col, row, t;
  int cont;

  rowSize = n/32;
  if (n%32)
    rowSize += 1;
  
  if (!(AT = (long *)malloc(64*rowSize*sizeof(long)))) {
    printf("doColumnOps(): Memory allocation error!\n");
    return -1;
  }
  if (!(BT = (long *)malloc(64*rowSize*sizeof(long)))) {
    printf("doColumnOps(): Memory allocation error!\n");
    free(AT);
    return -1;
  }
  /* Compute AT = A^T and BT = B^T. */
  memset(AT, 0x00, 64*rowSize*sizeof(long));
  memset(BT, 0x00, 64*rowSize*sizeof(long));

  for (row=0; row<n; row++) {
    for (col=0; col<32; col++) {
      if (A[2*row]&BIT(col&0x1F))
        AT[col*rowSize + row/32] ^= BIT(row&0x1F);
      if (B[2*row]&BIT(col&0x1F))
        BT[col*rowSize + row/32] ^= BIT(row&0x1F);
    }
    for (col=32; col<64; col++) {
      if (A[2*row + 1]&BIT(col&0x1F))
        AT[col*rowSize + row/32] ^= BIT(row&0x1F);
      if (B[2*row + 1]&BIT(col&0x1F))
        BT[col*rowSize + row/32] ^= BIT(row&0x1F);
    }
  }

  /* Now, apply row ops to get 'BT' into RREF. */
  col=0;
  for (row=0; row<64; row++) {
    /* Is there a row at or after 'row' with a '1' in column 'col'? */
    /* Consider adding a short circuit, testing whole words at a time */
    /* since we are generally expecting small rank for our matrices.  */
    cont=1;
    do {
      k=row;
      while ((k<64)&&((BT[rowSize*k + col/32]&BIT(col&0x1F))==0))
        k++;
      if (k < 64) {
        /* Swap rows 'row' and 'k'. */
        for (i=0; i<rowSize; i++) {
          t = BT[rowSize*row + i]; 
          BT[rowSize*row + i] = BT[rowSize*k + i];
          BT[rowSize*k + i] = t;
          t = AT[rowSize*row + i]; 
          AT[rowSize*row + i] = AT[rowSize*k + i];
          AT[rowSize*k + i] = t;
        }
        /* Use row 'row' to zero out the rest of this column. */
        for (i=0; i<64; i++) {
          if ((BT[i*rowSize + col/32]&BIT(col&0x1F)) && (i != row)) {
            for (j=0; j<rowSize; j++) {
              BT[i*rowSize + j] ^= BT[row*rowSize + j];
              AT[i*rowSize + j] ^= AT[row*rowSize + j];
            }
          }
        } 
        cont=0;
      } else {
        /* Column 'col' is zero at and below row 'row'. */
        cont=1;
      }
      col++;
    } while ((cont) && (col < n));
    if (col >= n)
      break; /* CJM, 10/20/03. */
  }

  /* Compute A = AT^T and B = BT^T. */
  for (i=0; i<2*n; i++)
    A[i] = B[i] = 0;

  for (row=0; row<64; row++) {
    for (col=0; col<n; col++) {
      if (AT[row*rowSize + col/32]&BIT(col&0x1F))
        A[2*col + row/32] ^= BIT(row&0x1F);
      if (BT[row*rowSize + col/32]&BIT(col&0x1F))
        B[2*col + row/32] ^= BIT(row&0x1F);
    }
  }
  free(AT);
  free(BT);
  return 0;
}

/*******************************************************/
int doColumnOps32(long *A, long *B, long n)
/*******************************************************/
/* Do column operations (Gaussian elimination) on 'B', */
/* applying the same operations to 'A' as we go.       */
/* Input: 'A' and 'B' are nx32 matrices.               */
/* Output: 'A' and 'B' are changed, as above. i.e., 'B'*/
/*         is in column-reduced echelon form, and the  */
/*         same transformation was applied to 'A'.     */
/*******************************************************/
/* Remark: This is horribly inefficient. There is      */
/* surely a better way to do it. But the easiest is to */
/* just transpose the matrices before and after.       */
/* But it doesn't much matter either way - this guy is */
/* not called from inside any loops, and it's not a    */
/* bottleneck.                                         */
/*******************************************************/
{ long *AT, *BT;
  long i, j, k, rowSize, col, row, t;
  int cont;

  rowSize = n/32;
  if (n%32)
    rowSize += 1;
  
  if (!(AT = (long *)malloc(32*rowSize*sizeof(long)))) {
    printf("doColumnOps(): Memory allocation error!\n");
    return -1;
  }
  if (!(BT = (long *)malloc(32*rowSize*sizeof(long)))) {
    printf("doColumnOps(): Memory allocation error!\n");
    free(AT);
    return -1;
  }
  /* Compute AT = A^T and BT = B^T. */
  memset(AT, 0x00, 32*rowSize*sizeof(long));
  memset(BT, 0x00, 32*rowSize*sizeof(long));

  for (row=0; row<n; row++) {
    for (col=0; col<32; col++) {
      if (A[row]&BIT(col&0x1F))
        AT[col*rowSize + row/32] ^= BIT(row&0x1F);
      if (B[row]&BIT(col&0x1F))
        BT[col*rowSize + row/32] ^= BIT(row&0x1F);
    }
  }

  /* Now, apply row ops to get 'BT' into RREF. */
  col=0;
  for (row=0; row<32; row++) {
    /* Is there a row at or after 'row' with a '1' in column 'col'? */
    /* Consider adding a short circuit, testing whole words at a time */
    /* since we are generally expecting small rank for our matrices.  */
    cont=1;
    do {
      k=row;
      while ((k<32)&&((BT[rowSize*k + col/32]&BIT(col&0x1F))==0))
        k++;
      if (k < 32) {
        /* Swap rows 'row' and 'k'. */
        for (i=0; i<rowSize; i++) {
          t = BT[rowSize*row + i]; 
          BT[rowSize*row + i] = BT[rowSize*k + i];
          BT[rowSize*k + i] = t;
          t = AT[rowSize*row + i]; 
          AT[rowSize*row + i] = AT[rowSize*k + i];
          AT[rowSize*k + i] = t;
        }
        /* Use row 'row' to zero out the rest of this column. */
        for (i=0; i<32; i++) {
          if ((BT[i*rowSize + col/32]&BIT(col&0x1F)) && (i != row)) {
            for (j=0; j<rowSize; j++) {
              BT[i*rowSize + j] ^= BT[row*rowSize + j];
              AT[i*rowSize + j] ^= AT[row*rowSize + j];
            }
          }
        } 
        cont=0;
      } else {
        /* Column 'col' is zero at and below row 'row'. */
        cont=1;
      }
      col++;
    } while ((cont) && (col < n));
    if (col >= n)
      break; /* CJM, 10/20/03. */
  }
  /* Compute A = AT^T and B = BT^T. */
  for (i=0; i<n; i++)
    A[i] = B[i] = 0;

  for (row=0; row<32; row++) {
    for (col=0; col<n; col++) {
      if (AT[row*rowSize + col/32]&BIT(col&0x1F))
        A[col] ^= BIT(row);
      if (BT[row*rowSize + col/32]&BIT(col&0x1F))
        B[col] ^= BIT(row);
    }
  }
  free(AT);
  free(BT);
  return 0;
}




