/**************************************************************/
/* combparts.c                                                */
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* Routines for combining partials to form full relations.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#ifndef _MSC_VER        
#include <sys/time.h>
#endif

#include "ggnfs.h"

#define LARGE_BUFFER 200000
#define FIELDSIZE(_L, _i) (_L->index[(_i)+1] - _L->index[_i])

/* Globals: */
unsigned char *counter=NULL;
long counterSize;

long mkUniqueLongs(long *x, long size);

/*************************************************************/
void sortByNumLP(long *fieldsBySize, llist_t *P, char *str)
/*************************************************************/
{ long i, size;
  long numBySize[8];

  ll_getsortOnFieldSize(fieldsBySize, P);
  if (str != NULL) {
    printf(str);
    for (i=0; i<8; i++) /* Overkill - 4 or 6 would do. */
      numBySize[i]=0;
    i=0;
    while (i<P->numFields) {
      if (fieldsBySize[i] > P->numFields) 
        printf("?!?WTF?!? fieldsBySize[%ld]=%ld vs. P->numFields=%ld.\n",
               i,fieldsBySize[i], P->numFields);
      size = P->index[fieldsBySize[i]+1] - P->index[fieldsBySize[i]];
      if (size > 6) 
        printf("?!Error!? Relation has more than 6 large primes!\n");
      else if (size < 0) {
        printf("P has been corrupted!\n");
        printf("P->numFields = %ld, P->index[%ld]=%ld, P->index[%ld]=%ld.\n",
               P->numFields, fieldsBySize[i], P->index[fieldsBySize[i]],
               fieldsBySize[i]+1, P->index[fieldsBySize[i]+1]);
      } else {
        numBySize[size] += 1;
      }
      i++;
    }
    for (i=0; i<5; i++)
      printf("There are %ld relations with %ld large primes.\n", numBySize[i], i);
  }
}

/*************************************************************/
long getMaxEntry(llist_t *P)
/*************************************************************/
{ long size, i, N=0;

  size = P->index[P->numFields];
if (size > P->maxDataSize) {
  printf("Error: P is broken!\n");
}
  for (i=0; i<size; i++)
    if (P->data[i] > N) N=P->data[i];
  return N;
}
/*************************************************************/
llist_t *mkLT(llist_t *P, long P0, long P1)
/*************************************************************/
/* Make a reverse-lookup table for entries of P in the range */
/* [P0, P1].                                                 */
/*************************************************************/
{ llist_t *L;
  long     i, j, size, cum, t, loc, pnum;

  /* Make one quick pass through to see how large the table will be,
     and how large each entry will be.
  */
  if (!(L = (llist_t *)malloc(sizeof(llist_t)))) {
    fprintf(stderr, "mkLT() memory allocation error!\n");
    exit(-1);
  }
  L->maxFields = P1-P0+1;
  if (!(L->index = (long *)calloc((5+L->maxFields), sizeof(long)))) {
    fprintf(stderr, "mkLT() memory allocation error!\n");
    exit(-1);
  }
    
  size = 0;
  for (i=0; i<P->numFields; i++) {
    for (j=P->index[i]; j<P->index[i+1]; j++) {
      pnum = P->data[j];
      if ((pnum>=P0) && (pnum<=P1)) {
        L->index[P->data[j]-P0] += 1;
        size++;
      }
    }
  }
  if (size == 0)
    printf("mkLT: There appear to be no large primes in the specified range.\n");

  L->maxDataSize = size+10;
  if (!(L->data = (long *)malloc(L->maxDataSize*sizeof(long)))) {
    fprintf(stderr, "mkLT() Memory allocation error!\n");
    exit(-1);
  }
  /* Now each L->index field has the number of relations that
     will be asociated with that prime. Go through and make
     it cumulative, so it will point to the start of relations
     for the corresponding prime.
  */
  size = P1-P0+1;
  cum=0;
  for (i=0; i<size; i++) {
    t = L->index[i];
    L->index[i] = cum;
    cum += t;
  }
  L->index[size] = cum; /* terminator. */
  L->numFields = size;

  /* Initialize the fields to all hold 0xFFFFFF, so we can tell where
     we have put stuff and where we haven't.
  */
  memset(L->data, 0xFF, L->maxDataSize*sizeof(long));

  /* Finally, put the data where it belongs. */
  for (i=0; i<P->numFields; i++) {
    for (j=P->index[i]; j<P->index[i+1]; j++) {
      pnum = P->data[j];
      if (pnum>=P0 && pnum<=P1) {
        loc = L->index[pnum-P0];
        while (L->data[loc] != 0xFFFFFFFF)
          loc++;
        L->data[loc] = i;
      }
    }
  }
  return L;
}



/*************************************************************/
int removeLPSingletons(llist_t *R, llist_t *P)
/*************************************************************/
/* Remove relations having a large prime which does not      */
/* appear in any of the other relations. This is done as     */
/* follows: (1) Determine an upper bound for the number of   */
/* large primes appearing.                                   */
/* (2) Make an array having 4 bits for each possible large   */
/*     prime.                                                */
/* (3) Make a pass through all relations. For each large     */
/*     prime encountered, set the corresponding counter to   */
/*     MIN(15, counter+1).                                   */
/* Finally, make one more pass through the data again looking*/
/* at the large primes. Any relations having a large prime   */
/* whose counter is == 1 will be deleted.                    */
/* So the time cost is about 2*(number of relations), and    */
/* the space cost is (max. # of large primes)/2 bytes.       */
/*  Upon completion, 'P' will be reallocated to its new size */
/* plus LARGE_BUFFER.                                        */
/*  The counter field is a global variable, as it will be    */
/* useful to some other functions. If anybody else ever      */
/* `free's it, they had better set it back to NULL as well!  */
/*************************************************************/
{ long     Np, i, j, h, size, ct;
  long     *removeList, rLSize, numRemove;
  long     *tmpPtr;
  unsigned char add[]={0x01, 0x10};


  Np=getMaxEntry(P)+4; /* a little padding, just in case. */
  /* We need four bits per prime in the hash table: */
  counterSize = 1 + Np/2;
  if (counter != NULL) free(counter);
  counter = (unsigned char *)calloc(counterSize,sizeof(unsigned char));

  size = P->index[P->numFields];
  for (i=0; i<size; i++) {
    h = P->data[i];
    ct = (counter[h/2]>>(4*(h%2))) & 0x0F;
    if (ct < 15) { 
      counter[h/2] += add[h&0x01]; 
    }
  }

  numRemove=0;
  rLSize = 32768;
  removeList = (long *)malloc(rLSize*sizeof(long));
  /* Now make a pass through looking for primes which appeared only once. */
  for (j=0; j<P->numFields; j++) {
    for (i=P->index[j]; i<P->index[j+1]; i++) {
      h = P->data[i];
      ct = (counter[h/2]>>(4*(h%2))) & 0x0F;
      if (ct == 1) {
        removeList[numRemove++] = j;
        if (numRemove >= (rLSize-1)) {
          rLSize += 32768;
          tmpPtr = realloc(removeList, rLSize*sizeof(long));
          if (tmpPtr==NULL) {
            fprintf(stderr, "removeLPSingletons() Memory (re-)allocation error!\n");
            exit(-1);
          }
          removeList = tmpPtr;
        }
        break;
      }
    }
  }
  /* 'removeList' should already be sorted and unique, but whatever. */
  numRemove = mkUniqueLongs(removeList, numRemove);
  printf("Deleting %ld singleton large primes.\n", numRemove);

  ll_deleteFields(P, removeList, numRemove);
  ll_resize(P, P->index[P->numFields] + LARGE_BUFFER);
  ll_deleteFields(R, removeList, numRemove);
//  printf("There are %ld relations remaining.\n", R->numFields);
  free(removeList);
  return (int)numRemove;
}

/*************************************************************/
long mkUniqueLongs(long *x, long size)
/*************************************************************/
{ long i, j;

  if (size==0) return 0;
  qsort(x, size, sizeof(long), cmpLongs);
  for (i=0, j=1; j<size; j++) 
    if (x[j] != x[i])
      x[++i] = x[j];
  return i+1;
}

/*************************************************************/
long mkUniquePairs(lpair_t *pairs, long numPairs)
/*************************************************************/
/* Make sure there are no two pairs with the same x value.   */
/*************************************************************/
{ long i, j, k;

  if (numPairs==0) return 0;
  qsort(pairs, numPairs, sizeof(lpair_t), cmp_lpair_t);
  k=0;
  for (j=0,i=1; i<numPairs; i++) {
    if (pairs[i].x != pairs[j].x) {
      pairs[++j].x = pairs[i].x;
      pairs[j].y = pairs[i].y;
    }
  }
  return j+1;
}



#define MAX_MERGE_OPS 500000
/*************************************************************/
int merge(llist_t *R, llist_t *P, llist_t *revP, long P0, long P1, int level)
/*************************************************************/
/* This does the actual combining of relations to eliminate  */
/* unbalanced large primes.                                  */
/*************************************************************/
{ long i, j, k, pnum, relnum, numAdds, c1, ck;
  long rwt[64], pwt[64], minrwt, minpwt, pivot;
  lpair_t *pairs;

  pairs = (lpair_t *)calloc(MAX_MERGE_OPS, sizeof(lpair_t));
  numAdds=0;
  for (i=0; (i<revP->numFields)&&(numAdds < MAX_MERGE_OPS); i++) {
    pnum = P0+i;
    k = FIELDSIZE(revP, i);
    if (k>1) {
      /* This large prime appears in k relation-sets. */
      k = MIN(k, 64); /* k should never be anywhere near 64! */
      /* How many relations are contributing to each relation set? */
      minrwt=64;
      for (j=0; j<k; j++) {
        relnum = revP->data[revP->index[i]+j];
        rwt[j] = FIELDSIZE(R, relnum);
        minrwt = MIN(minrwt, rwt[j]);
      }

      /* How many large primes are there in each relation set? */
      minpwt=64;
      for (j=0; j<k; j++) {
        relnum = revP->data[revP->index[i]+j];
        pwt[j] = FIELDSIZE(P, relnum);
        minpwt = MIN(minpwt, pwt[j]);
      }
      pivot=-1;
      for (j=0; j<k; j++) {
        if ((pwt[j]==minpwt) && ((pivot==-1) || (rwt[j]<rwt[pivot])))
          pivot = j;
      }
      if (pivot >= 0) {
        /* Check to see if we can use it. */
        if (minpwt > 1) pivot=-1;
        else if (rwt[pivot] > level) pivot=-1;
      }
      if (pivot >= 0) {
        ck = revP->data[revP->index[i]+pivot];
        /* Ok - see what additions we can do. */
        for (j=0; j<k; j++) {
          if (j != pivot) {
            c1 = revP->data[revP->index[i]+j];
            if ((rwt[pivot]+rwt[j] <= level)&&(numAdds<MAX_MERGE_OPS)) {
              pairs[numAdds].x = c1; pairs[numAdds++].y = ck;
            }
          }
        }
      }
    }
  }
  numAdds = mkUniquePairs(pairs, numAdds);
  if (numAdds > 0) {
    printf("Doing %ld additions...\n", numAdds);
    ll_catFields(P, pairs, numAdds, 1);
    ll_catFields(R, pairs, numAdds, 1);
  }
  free(pairs);
  return 0;
}


/* The maximum # of `longs' we can allocate for a reverse-lookup
   table. The larger this is, the more memory makePass will use,
   but it will also be faster. When I'm less tired, I'll figure
   the runtime and put a more precise comment here.
*/
#define MAX_PART_DATASIZE 2500000
/*************************************************************/
long makePass(llist_t *R, llist_t *P)
/*************************************************************/
{ long *fieldsBySize, startNumLP[7];
  long  i, k, bd, pMax, numParts, part, partSize, P0, P1;
  long  lastSize, numFulls, s0, s1;
  llist_t *revP;

  /* Remove any singletons. But this function won't take the liberty
     of re-running itself after it removes the singletons. As a result,
     removing some singletons could create some more. So repeat until
     it's stable.
       This will also set up `counter', which counts the number of
     times a given large prime appears in some relation-set. 
  */
if (!(fieldsBySize = (long *)malloc(P->numFields*sizeof(long)))) {
  fprintf(stderr, "makePass() Severe memory allocation error!\n");
  exit(-1);
}
  printf("Before sortByNumLP()... Doing ll_verify(P)...\n");
  if (ll_verify(P)) {
    printf("ll_verify() reported an error!\n");
    exit(-1);
  }
  printf("ll_verify() reports that 'P' appears to be intact.\n");
sortByNumLP(fieldsBySize, P, "makePass:\n");
  printf("After sortByNumLP()... Doing ll_verify(P)...\n");
  if (ll_verify(P)) {
    printf("ll_verify() reported an error!\n");
    exit(-1);
  }
  printf("ll_verify() reports that 'P' appears to be intact.\n");
free(fieldsBySize);

  s0 = R->numFields;
  do {
    lastSize = P->numFields;
    removeLPSingletons(R, P); 
  } while (P->numFields < lastSize);
  s1 = R->numFields;
  printf("%ld singletons deleted.\n", s0-s1);

  /* Sort the relation-sets on the number of large primes
     appearing. We don't actually sort 'P', but rather obtain
     a list of fields of P which is sorted in ascending order.
  */
  if (!(fieldsBySize = (long *)malloc(P->numFields*sizeof(long)))) {
    fprintf(stderr, "makePass() Severe memory allocation error!\n");
    exit(-1);
  }
  sortByNumLP(fieldsBySize, P, "makePass:\n");
  /* Find the start of fields with 1 LP, 2 LP,... */
  startNumLP[0]=0;
  bd = P->numFields;
  for (i=1; i<=4; i++) {
    k=startNumLP[i-1];
    while ((k < bd) && (FIELDSIZE(P,fieldsBySize[k])<i))
      k++;
    startNumLP[i]=k;
  }

  /*****************************************************************/
  /* Now do as follows:                                            */
  /* (1) Find the maximum large prime appearing, maxP.             */
  /* (2) Partition the possible primes [0, maxP] into subsets      */
  /*     of manageable size.                                       */
  /* (3) For each partition of the primes, create a reverse-lookup */
  /*     table, so we can look the prime up to see in which        */
  /*     relation-sets it appears.                                 */
  /* (4) Use the reverse-lookup and counter tables to decide what  */
  /*     column additions to make, if any.                         */
  /*****************************************************************/
  pMax=getMaxEntry(P);
  numParts = 1 + P->index[P->numFields]/MAX_PART_DATASIZE;
  P0 = 0;
  P1 = pMax/numParts;
  partSize = P1+1;
  part=0;
  do {
    /* Make the reverse-lookup table. */
    printf("Doing merge on chunk %ld/%ld  (P0=%ld, P1=%ld)...\n", part+1, numParts, P0, P1);
    revP = mkLT(P, P0, P1);
    merge(R, P, revP, P0, P1, MAX_RELS_IN_FF);

    /* Free up the reverse-lookup table. */
    ll_clear(revP); free(revP);
    if (++part < numParts) {
      P0 = P1+1;
      P1 = MIN(P0+partSize, pMax);
    }
  } while (part < numParts);
  free(fieldsBySize);
  /* Count the number of full relations: */
  numFulls=0;
  for (i=0; i<P->numFields; i++)
    if (FIELDSIZE(P, i)==0)
      numFulls++;
  return numFulls;
}

/*************************************************************/
long combParts(llist_t *R, llist_t *P)
/*************************************************************/
/*Input:
     'P' is a list containing the large primes. It should not
     actually contain the primes, but rather a unique integer associated
     with each large prime (i.e., algebraic and rational primes have
     disjoint sets of indicies).
     On input, it should be indexed by relation number.
     'R' should point to an uninitialized llist_t.
  Output:
     'R' is a list of full relations. Each field contains
     a list of relation numbers contributing to the 
     corresponding full relation. In Cavallar's nomenclature, 'R'
     will be a collection of relation sets.
     'P' will be modified in the process.
  Return value: The number of full relations built.

  This will be a rather memory intensive function, so be sure to
  free up any unneeded memory before calling us! Everything will
  be done in RAM instead of with disk I/O. So, for example, with
  50M relations having an average of 3 large primes each, you will
  be looking at a 'P' structure with 800MB (assuming 4 byte longs).
  Then, tack onto that the RAM for the 'R' structure and miscellaneous,
  and you could easily be around 1.5BGB of RAM! 

  This function is unwritten at the moment - I want to put a little
  thought into it so I get it right this time and get something near
  an optimal algorithm which is at the same time efficient.
*/
/**************************************************************/ 
{ long i;
  long lastFull, full;
  int  pass=0;

  /* We can reallocate if necessary. So if this seems to be way
     too much memory, decrease it at will.
  */
  printf("combParts() Doing ll_verify(P)...\n");
  if (ll_verify(P)) {
    printf("ll_verify() reported an error!\n");
    exit(-1);
  }
  printf("ll_verify() reports that 'P' appears to be intact.\n");


  if (ll_init(R, P->maxDataSize/4, P->maxFields)) {
    fprintf(stderr, "combParts: ll_init() reports severe error!\n");
    exit(-1);
  }
  for (i=0; i<P->numFields; i++) {
    ll_appendField(R, &i, 1);
  }

  full = 0;
  do {
    printf("** pass %d...\n", ++pass);
    lastFull = full;
    full = makePass(R, P);
    printf("* There are now %ld full relations.\n", full);
  } while (lastFull < full);


  return full;
}
