/**************************************************************/
/* getdeps.c                                                  */
/* Copyright 2004, Chris Monico.                              */
/* Many fixes and improvements due to KAMADA Makoto.          */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>



#ifndef _MSC_VER 
#include <sys/time.h>
#endif
#include "ggnfs.h"


/* I need to figure out what is roughly optimal
   for a machine with, say 512MB of RAM. This is
   the max file size for individual processed relation
   files. But beware - don't make it too small - some
   of the internal algorithms may be
   quadratic in the number of files! On the other hand,
   if this is too large, you'll run into VM and everything
   will slow to a crawl or outright crash.
    Anyway, if you run into RAM problems running this program,
   this is something you can try to decrease. Just don't get
   carried away.
*/
#define DEFAULT_MAX_FILESIZE 200000000

#define MAX_MEM_ALLOC 128000000
#define MAX_SPAIRS_ALLOC 4000000
#define DEFAULT_QCB_SIZE 62
#define DEFAULT_SEED 1
#define DEFAULT_NUM_FILES 4
#define TMP_FILE "tmpdata.000"
#define DEFAULT_DEPNAME "deps"
#define DEFAULT_COLNAME "cols"
#define DEFAULT_LPI_NAME "lpindex"

#define _BINARY

#define USAGE " -fb <fname> -prel <fname> -newrel <fname> [-qs <qcb size>] [-v]\n"\
"-fb <fname>           : Factor base.\n"\
"-prel <file prefix>   : File name prefix for input/output of processed relations.\n"\
"-newrel <fname>       : File name for new, unprocessed relations.\n"\
"-nodfactor            : Don't try to factor the discriminant again.\n"\
"-forceCC              : Force a cycle count, even if there are too many large primes.\n"

#define START_MSG \
"\n"\
" __________________________________________________________ \n"\
"|        This is the getdeps program for GGNFS.            |\n"\
"| Version: %-15s                                 |\n"\
"| This program is copyright 2004, Chris Monico, and subject|\n"\
"| to the terms of the GNU General Public License version 2.|\n"\
"|__________________________________________________________|\n"


typedef struct {
  long  a, b;     /* The relation is (a - b*alpha). */
  long  relNum;   /* The relation number, within the corresponding rel_list structure. */
} rel_t;

typedef struct {
  long p;
  long r;
} af_t;

/***********************************************************/
/* This structure is for columns that are being processed. */
/* Before processing, we know only the relations that      */
/* comprise each column. After processing, we know all the */
/* nonzero rows of the matrix (i.e., the primes appearing  */
/* with odd exponent and the QCB (sign bit is in QCB)).    */
/* During processing, we will know some of each. That is,  */
/* we might have already replaced some relations with their*/
/* factorizations, but not yet all of them. These data are */
/* what will be stored in the file cols.np.                */
/***********************************************************/
typedef struct {
  char numRels;
  int  numPrimes;
  long Rels[MAX_RELS_IN_FF];
  long QCB[2];
  long rows[MAX_ROWS_IN_COL];
} column_t;




/***** Globals *****/
int verbose=0, discFact=1, forceCC=0;
long initialFF=0, initialRelations=0, finalFF=0;

long makeMasterIndex2(multi_file_t *lpF, multi_file_t *relF, long estN);

/******************************************************/
void writeColIndex(FILE *fp, column_t *C)
/******************************************************/
/* This is the indicies of the relations contributing */
/* to the column. We don't care about anything else.  */
/******************************************************/
{ 
  fwrite(&C->numRels, sizeof(char), 1, fp);
  if (C->numRels > 0)
    fwrite(&C->Rels, sizeof(long), C->numRels, fp);
}
/******************************************************/
void readColIndex(column_t *C, FILE *fp)
/******************************************************/
/* This is the indicies of the relations contributing */
/* to the column. We don't care about anything else.  */
/******************************************************/
{ 
  fread(&C->numRels, sizeof(char), 1, fp);
  if (C->numRels > 0)
    fread(&C->Rels, sizeof(long), C->numRels, fp);
}

/******************************************************/
void writeColLF(FILE *fp, column_t *C)
/******************************************************/
{ 
  fwrite(&C->numRels, sizeof(char), 1, fp);
  fwrite(&C->numPrimes, sizeof(int), 1, fp);
  if (C->numRels > 0)
    fwrite(&C->Rels, sizeof(long), C->numRels, fp);
  fwrite(&C->QCB, sizeof(long), 2, fp);
  fwrite(&C->rows, sizeof(long), C->numPrimes, fp);
}

/******************************************************/
int readColLF(column_t *C, FILE *fp)
/******************************************************/
{ 
  fread(&C->numRels, sizeof(char), 1, fp);
  fread(&C->numPrimes, sizeof(int), 1, fp);
  if (C->numRels > 0)
    fread(&C->Rels, sizeof(long), C->numRels, fp);
  fread(&C->QCB, sizeof(long), 2, fp);
  if (feof(fp)) return -1;
  fread(&C->rows, sizeof(long), C->numPrimes, fp);
  return 0;
}

/******************************************************/
void writeColSF(FILE *fp, column_t *C)
/******************************************************/
/* For these, all we care about is the prime-indexing */
/* fields and nothing else.                           */
/******************************************************/
{ 
  fwrite(&C->numPrimes, sizeof(int), 1, fp);
  fwrite(&C->QCB, sizeof(long), 2, fp);
  fwrite(&C->rows, sizeof(long), C->numPrimes, fp);
}

/******************************************************/
void readColSF(column_t *C, FILE *fp)
/******************************************************/
{ 
  fread(&C->numPrimes, sizeof(int), 1, fp);
  fread(&C->QCB, sizeof(long), 2, fp);
  fread(&C->rows, sizeof(long), C->numPrimes, fp);
}


/******************************************************/
int cmpRel_t_ab(const void *x, const void *y)
/******************************************************/
{ rel_t *X = (rel_t *)x, *Y=(rel_t *)y;

  if (X->b < Y->b) return -1;
  if (X->b > Y->b) return 1;
  if (X->a < Y->a) return -1;
  if (X->a > Y->a) return 1;
  return 0;
}
  
/****************************************************/
int prepRelFilter(rel_t *Rp, rel_list *RL)
/****************************************************/
{ long i, index;

  for (i=0; i<RL->numRels; i++) {
    index = RL->relIndex[i];
    /* Now, recall: the first data element is the size field. The next two
       will be the a, b fields respectively.
    */
    Rp[i].a = RL->relData[index+1]; 
    Rp[i].b = RL->relData[index+2];
    Rp[i].relNum = i;
  }
  return 0;    
}

/*************************************************************/
long filterRelations(rel_t *Rp, multi_file_t *relF, int fileNum, rel_list *RL)
/*************************************************************/
/* Filter out duplicates using the provided Rp structure.    */
/* 'Rp' must have already been allocated and put through     */
/* prepRelFilter().                                          */
/*************************************************************/
/* We should consider doing this differently, with a hash    */
/* table and collision detection to avoid sorting altogether.*/
/*************************************************************/
{ long   i, uniq;
  FILE  *fp;
  char   fName[64];

  if (RL->numRels <= 0) {
    fprintf(stderr, "filterRelations() Called with no valid relations!\n");
    return 0;
  }
  sprintf(fName, "%s.%d", relF->prefix, fileNum);
  /* Sort them on the (a,b) component, so we can find
     duplicates. Note: we use reverse lex, since it may
     be useful to see how many relations there are with
     a particular 'b' component. If there are too few, we
     might then consider doing some linear sieving over
     those 'b' values.
     We would probably do well to use a different sorting
     algorithm actually. The problem is that it will be
     quite possible that a large portion of the list is
     already sorted (i.e., the relations that were processed
     some time in the past and just read in from file).
     I seem to recall that quick sort is not particularly
     good in this case. If so, this will kill us when we
     have large data sets. But it's ok for now.
  */
  qsort(Rp, RL->numRels, sizeof(rel_t), cmpRel_t_ab);
  
  /* Now, a linear pass through the array to remove duplicates. */
  for (i=1, uniq=1; i<RL->numRels; i++) {
    /* Compare 'a' first, since many of the 'b' fields will be equal! */
    if ((Rp[i].a != Rp[uniq-1].a) || (Rp[i].b != Rp[uniq-1].b)) {
      Rp[uniq].a = Rp[i].a;     
      Rp[uniq].b = Rp[i].b;
      Rp[uniq].relNum = Rp[i].relNum;
      uniq++;
    }
  }
  printf("Filtering found %ld unique relations.\n", uniq);
  
  /****************************************************************/
  /* Now, sort and filter out the duplicates from the rel_list.   */
  /* Since the elements of rel_list have variable size, this      */
  /* gets tricky. Probably the safest way is to write them back   */
  /* to disk correctly, and then read them back (i.e., they       */
  /* have to be written back to disk eventually anyway, so we     */
  /* only waste time reading them back in. We can live with that. */
  /****************************************************************/
  if (!(fp = fopen(fName, "wb"))) {
    fprintf(stderr, "Error opening %s for write!\n", fName);
    exit(-1);
  }
  writeRawLong(fp, &uniq);
  printf("Writing %ld relations:\n", uniq);
  for (i=0; i<uniq; i++) {
    writeRel(fp, &RL->relData[RL->relIndex[Rp[i].relNum]]);
  }
  fclose(fp);
  printf("%ld unique relations written back to %s.\n", uniq, fName);

  /* Now, we probably have way more memory allocated then
     we actually need, so we could do a realloc - but we'll
     leave that up to the caller, if they so desire. i.e.,
     maybe they allocated enough to handle the largest file,
     instead of doing malloc,free,malloc,free,...
  */
  if (readRelList(RL, fName)) {
    fprintf(stderr, "Some error occurred reading back the relations from %s!\n", fName);
    printf("RL->maxDataSize = %ld.\n", RL->maxDataSize);
    exit(-1);
  }
  return uniq;
}

/*****************************************************/
void appendToLFile(char *fName, long p, long r)
{ FILE *fp;
  long _p=p, _r=r, z=0; 

  if (!(fp = fopen(fName, "ab")))
    return;
  writeRawLong(fp, &_p); 
  writeRawLong(fp, &_r); 
  writeRawLong(fp, &z);
  writeRawLong(fp, &z);
  fclose(fp);
}



typedef struct {
  long p;
  long r;
  long index;
  long counter;
} mmi_table_t;
/*******************************************************/
int cmpMMI_t(const void *A, const void *B)
/*******************************************************/
{ mmi_table_t *a = (mmi_table_t *)A, *b=(mmi_table_t *)B;

  if (a->p < b->p) return -1;
  if (a->p > b->p) return 1;
  if (a->r < b->r) return -1;
  if (a->r > b->r) return 1;
  return 0;
}

/**************************************************************************/
long makeMasterIndex(multi_file_t *lpF, multi_file_t *relF, long estN)
/**************************************************************************/
/* lpFPrefix is the filename prefix where the indices are to be stored.   */
/* relPRefix is the filename prefix for the `numRelFiles' relation files. */
/* estN is an estimate for the number of large primes. For this estimate, */
/* it's better to err on the high side than the low side.                 */
/* Return value: The number of large primes found, or -1 on error.        */
/**************************************************************************/
/* It's assumed that the filenames are reasonably sized (< 60 bytes).     */
/**************************************************************************/
{ long         i, j, k=0, K, Tsize, h, N=0, p, r, entry, numRelsInFile, hashSize;
  long         _p, _r, _h, _ct, _index;
  mmi_table_t *T;
  char         fName[64], LfName[64];
  FILE        *fp;
  int          res=0, alg;
  relation_t   R;

  hashSize = 4*getNextPrime(estN);
  K = sizeof(mmi_table_t)*hashSize/MAX_MEM_ALLOC + 1;
  lpF->numFiles = K;
  Tsize = (hashSize/K);
  lpF->data1 = Tsize;
  lpF->data2 = hashSize;
  if (!(T = (mmi_table_t *)malloc(Tsize*sizeof(mmi_table_t)))) {
    fprintf(stderr, "makeMasterIndex() Error allocating %ldMB for the T table!\n",
            Tsize*sizeof(mmi_table_t)/1048576);
    exit(-1);
  }
  sprintf(LfName, "%s.L", lpF->prefix);
  if (!(fp = fopen(LfName, "wb"))) {
    fprintf(stderr, "makeMasterIndex(): Error opening file %s for write!\n", LfName);
    res = -1;
    goto MMI_SS_DONE;
  }
  fclose(fp);

  N=0;
  for (j=0; j<K; j++) {
    printf("Large prime pass %ld of %ld.\n", j+1, K);
    /* Initialize T: */
    for (i=0; i<Tsize; i++) {
      T[i].p = 0; T[i].r = -1;
      T[i].index = -1; T[i].counter = 0;
    }

    for (i=0; i<relF->numFiles; i++) {
      sprintf(fName, "%s.%ld", relF->prefix, i);
      if (!(fp = fopen(fName, "rb"))) {
        fprintf(stderr, "makeMasterIndex() Error opening relation file %s for read! Ignoring.\n", fName);
      } else {
        readRawLong(&numRelsInFile, fp);
        printf("Scanning %ld relations from %s...\n", numRelsInFile, fName);
        while (readRel(&R, fp)==0) {
          k=0; alg=0;
          while ((alg==0) || ((k < MAX_LARGE_ALG_PRIMES) && (R.a_p[k] > 1))) {
            /* Get the next large prime in this relation, be it rational or algebraic. */
            p=0; r=-1;
            if ((alg==0) && (k < MAX_LARGE_RAT_PRIMES) && (R.p[k]>1)) {
              p = R.p[k]; r=-1;
            } else if (alg==0) {
              alg=1; k=0;
            }
            if (alg && (k < MAX_LARGE_ALG_PRIMES) && (R.a_p[k] > 1)) {
              p = R.a_p[k]; r = R.a_r[k];
            }
            k++;
            if (p > 1) {
              /* OK - we found a large prime. First, compute its hash value. */
              h = NFS_HASH_Q(p, r, hashSize);
              
              /* Is the hash value in the proper range for this output file? */
              if ((h >= j*Tsize) && (h < (j+1)*Tsize)) {

                /* It is: retrieve the corresponding entry. */
                entry = h - j*Tsize;
                if ((T[entry].p == p) && (T[entry].r == r))
                  T[entry].counter += 1;
                else if (T[entry].p == 0) {
                  T[entry].p = p; T[entry].r = r; T[entry].counter = 1;
                } else {
                  /* This prime is not in the hash table, but something else is in       */
                  /* it's place. Just append this prime to the L file for later sorting. */
                  appendToLFile(LfName, p, r);
                }
              }
            }
          }
        }
        fclose(fp); /* All done with this relation file. */
      }
    }
    /* Set the index fields: */
    for (i=0; i<Tsize; i++) {
      if (T[i].p > 1) {
        T[i].index = N++;
      }
    }
    /* Finally, we should write the 'T' table out to file. */
    sprintf(fName, "%s.%ld", lpF->prefix, j);
    if (!(fp = fopen(fName, "wb"))) {
      fprintf(stderr, "makeMasterIndex(): Error opening file %s for write!\n", fName);
      res = -1;
      goto MMI_SS_DONE;
    }
    for (i=0; i<Tsize; i++) {
      if (T[i].p > 1) {
        /* Perhaps we might also want to restrict our attention to primes
           appearing more than once?
        */
        _p = T[i].p; _r = T[i].r; _h = i+(j*Tsize);
        _index = T[i].index; _ct = T[i].counter;
        writeRawLong(fp, &_h); writeRawLong(fp, &_p);
        writeRawLong(fp, &_r); writeRawLong(fp, &_index);
        writeRawLong(fp, &_ct);
      }
    }
    fclose(fp);
    
  }
MMI_SS_DONE:
  printf("makeMasterIndex(): Found %ld large primes + those in %s.\n", N, LfName);
  printf("Tsize = %ld\n", Tsize);
  /* Now, fix the L table. That is, sort it, remove duplicates (and keep count), */
  /* and give the elements consistent indices.                                   */
  if (!(fp = fopen(LfName, "rb"))) {
    printf("Warning: Could not open %s for read.\n", LfName);
  } else {
    i=0;
    while ((i<Tsize) && (!(feof(fp)))) {
      _p = 1;
      readRawLong(&_p, fp); 
      readRawLong(&_r, fp); 
      readRawLong(&_index, fp);
      readRawLong(&_ct, fp);
      if (_p>1) {
        T[i].p = _p; T[i].r = _r; T[i].counter = 1;
        i++;
      }   
    }
    fclose(fp);
    printf("Read %ld entries from %s, counting repetitions.\n", i, LfName);
    if (i >= Tsize) 
      printf("Warning: read Tsize = %ld large primes from %s - there may have been more!\n", Tsize, LfName);
    if (i<=1) {
      k=i; goto SS_UP_DONE;
    }
    qsort(T, i, sizeof(mmi_table_t), cmpMMI_t);
    /* Now, reduce to unique, counting occurrences. */
    j=1; k=0;
    while (j<i) {
      if ((T[j].p==T[k].p) && (T[j].r == T[k].r)) {
        T[k].counter += 1;
        j++;
      } else {
        k++;
        T[k].p = T[j].p; T[k].r = T[j].r; T[k].counter = 1;
        j++;
      }
    }
    k++;
SS_UP_DONE:
    /* Now, the first k entries: T[0],...,T[k-1] are unique. */
    if ((fp = fopen(LfName, "wb"))) {
      for (i=0; i<k; i++) {
        _p = T[i].p; _r = T[i].r; _index = N+i; _ct = T[i].counter;
        writeRawLong(fp, &_p); 
        writeRawLong(fp, &_r); 
        writeRawLong(fp, &_index);
        writeRawLong(fp, &_ct);
      }
      fclose(fp);
    }
  }
  N += k;
  printf("Found %ld large primes in total.\n", N);

  free(T);
  if (res)
    return res;
  return N;
}

/*********************************************************************/
void writeRowLF(FILE *fp, char byte, long *data, int dataSize)
/**********************************************************************/
/* During processing, while we look up the indicies for each large    */
/* prime, the 'byte' field signifies which primes have been converted.*/
/* i.e., accoriding to the bits.                                      */
/**********************************************************************/
{ char b=byte;
  fwrite(&b, sizeof(char), 1, fp);
  fwrite(data, sizeof(long), dataSize, fp);
}

/*********************************************************************/
int readRowLF(char *byte, long *data, int dataSize, FILE *fp)
/*********************************************************************/
{ 
  if (fread(byte, sizeof(char), 1, fp) != 1) return -1;
  if (fread(data, sizeof(long), dataSize, fp) < dataSize) return -1;
  return 0;
}

/*********************************************************************/
void writeRowSF(FILE *fp, long b1, long b2, long *data)
/**********************************************************************/
/* After processing, the first two bytes give (1) # of large primes.  */
/* (2) # of relations in this row.                                    */
/**********************************************************************/
{ long B1=b1, B2=b2;
  int  dataSize=(int)b1 + (int)b2;

  fwrite(&B1, sizeof(long), 1, fp);
  fwrite(&B2, sizeof(long), 1, fp);
  fwrite(data, sizeof(long), dataSize, fp);
}

/*********************************************************************/
int readRowSF(long *b1, long *b2, long *data, FILE *fp)
/*********************************************************************/
{ int dataSize, rSize=0;

  rSize += fread(b1, sizeof(long), 1, fp);
  rSize += fread(b2, sizeof(long), 1, fp);
  if (feof(fp)) return -1;
  dataSize = (int)(*b1 + *b2);
  rSize += fread(data, sizeof(long), dataSize, fp);
  if (((rSize-2) != dataSize)||(dataSize ==0)) return -1;
  return 0;
}


/*********************************************************************/
long loadLPrimes(mmi_table_t **Lprimes, multi_file_t *lpF)
/*********************************************************************/
/* Lprimes is not allocated - we will take care of allocation here.  */
/* Lprimes will be filled with the leftover large primes living in   */
/*   "lpF->prefix".L                                                 */
/* Return value: The number of primes read in.                       */
/*********************************************************************/
{ FILE   *fp;
  char   fName[64];
  struct stat fileInfo;
  long   maxLPrimes, p, r, ct, index, numLPrimes=0;
  mmi_table_t *tmpPtr;

  *Lprimes = NULL;
  sprintf(fName, "%s.L", lpF->prefix);
  if (stat(fName, &fileInfo)) 
    return 0;

  maxLPrimes = 100 + (fileInfo.st_size/sizeof(mmi_table_t));
  *Lprimes = (mmi_table_t *)malloc(maxLPrimes*sizeof(mmi_table_t));

  if (!(fp = fopen(fName, "rb"))) {
    free(*Lprimes); *Lprimes = NULL; return 0;
  }
  while (!(feof(fp))) {
    p=1;
    readRawLong(&p, fp); 
    readRawLong(&r, fp); 
    readRawLong(&index, fp);
    readRawLong(&ct, fp);
    if (p>1) {
      (*Lprimes)[numLPrimes].p = p;
      (*Lprimes)[numLPrimes].r = r;
      (*Lprimes)[numLPrimes].index = index;
      (*Lprimes)[numLPrimes].counter = ct;
      numLPrimes++;
    }
  }
  fclose(fp);
  tmpPtr = realloc(*Lprimes, numLPrimes*sizeof(mmi_table_t));
  if (tmpPtr != NULL) *Lprimes = tmpPtr;
  else {
    printf("loadLPrimes() memory reallocation error!\n");
  }
  return numLPrimes;
}

/*********************************************************************/
int loadTtable(mmi_table_t *T, long Tsize, multi_file_t *lpF, int fileNum)
/*********************************************************************/
/* Same as above, except that 'T' is already allocated and we store  */
/* the elements in 'T' indexed by their hash value.                  */
/*********************************************************************/
{ long  h, p, r, index, ct, hashOffset;
  long  h0, h1;
  char  fName[64];
  FILE *fp;

  
  hashOffset = fileNum*lpF->data1;
  h0 = fileNum*lpF->data1;
  h1 = (fileNum+1)*lpF->data1;

  sprintf(fName, "%s.%d", lpF->prefix, fileNum);
  if (!(fp = fopen(fName, "rb"))) {
    fprintf(stderr, "loadTtable() Could not open %s for read!\n", fName);
    return -1;
  }
  /* First, zero out the table. */
  for (h=0; h<Tsize; h++) {
    T[h].p = 0; T[h].r = -1;
    T[h].index = -1; T[h].counter = 0;
  }
  
  while (!(feof(fp))) {
    p=1;
    readRawLong(&h, fp); readRawLong(&p, fp);
    readRawLong(&r, fp); readRawLong(&index, fp);
    readRawLong(&ct, fp);
    /* Need to check that the hash is in the proper range,
       I think.
    */
    if (p>1) {
      T[h-hashOffset].p = p;
      T[h-hashOffset].r = r;
      T[h-hashOffset].index = index;
      T[h-hashOffset].counter = ct;
    }
  }
  fclose(fp);
  return 0;
}

  
/*********************************************************************/
long findLPrime(mmi_table_t *L, long tableSize, long p, long r)
/*********************************************************************/
/* If the given (p,r) prime is in the L table, return it's index.    */
/* Otherwise, return -1.                                             */
/*********************************************************************/
{ mmi_table_t *res, thisP;
  long index;

  thisP.p = p; thisP.r = r;  

  res = (mmi_table_t *)bsearch(&thisP, L, tableSize, sizeof(mmi_table_t), cmpMMI_t);
  if (res==NULL)
    return -1;
  index = (res - L);
  return index;
}

/*********************************************************************/
rel_list *getRelList(multi_file_t *prelF, int index)
/*********************************************************************/
/* Allocate for and read in the specified relation file. Caller is   */
/* obviously responsible for freeing the memory when done!           */
/*********************************************************************/
{ rel_list *RL;
  FILE     *fp;
  char      fName[256];
  struct stat fileInfo;

  RL = (rel_list *)malloc(sizeof(rel_list));
  RL->maxDataSize = 0;
  sprintf(fName, "%s.%d", prelF->prefix, index);
  
  if (stat(fName, &fileInfo)) {
    printf("Could not stat file %s!\n", fName);
    free(RL); return NULL;
  }
  RL->maxDataSize = fileInfo.st_size/sizeof(long);
  if ((fp = fopen(fName, "rb"))) {
    readRawLong(&RL->maxRels, fp);
    fclose(fp);
  }
  RL->maxRels += 1;
  /* Now allocate for the relations. */
  if (!(RL->relData = (long *)malloc(RL->maxDataSize * sizeof(long)))) {
    fprintf(stderr, "Error allocating %ldMB for reading relation list!\n",
            RL->maxDataSize * sizeof(long)/1048576);
    free(RL); return NULL;
  }
  if (!(RL->relIndex = (long *)malloc(RL->maxRels * sizeof(long)))) {
    fprintf(stderr, "Error allocating %ldMB for relation pointers!\n", 
            RL->maxRels * sizeof(long)/1048576);
    free(RL->relData); free(RL);
    return NULL;
  }
  RL->numRels = 0;
  readRelList(RL, fName);
  return RL;
}

/*********************************************************************/
void clear_rel_list(rel_list *RL)
{
  if (RL->relData != NULL) 
    free(RL->relData);
  if (RL->relIndex != NULL)
    free(RL->relIndex);
  RL->relData = RL->relIndex = NULL;
}
  

/*********************************************************************/
void makeRows(multi_file_t *rowF, multi_file_t *prelF, multi_file_t *lpF)
/*********************************************************************/
/* rowF is where the row info will go (i.e., which large primes      */
/* appear in a given row (by index)). A row may consist of the       */
/* the combination of several relations.                             */
/* prelF is where the processed relations are living.                */
/* lpF is where the (indexed) large primes are living.               */
/*********************************************************************/
/* Note: This is simplified by the fact that we know already how     */
/* many relations each prime appears in  - when we created the index,*/
/* we took note of this and recorded it along with the indexing data.*/
/*********************************************************************/
{ int        i, dataSize = MAX_LARGE_RAT_PRIMES + 2*MAX_LARGE_ALG_PRIMES;
  long       data[MAX_LARGE_RAT_PRIMES + 2*MAX_LARGE_ALG_PRIMES];
  long       newData[MAX_LARGE_RAT_PRIMES + MAX_LARGE_ALG_PRIMES];
  long       j, k, l, p, r, h, index, relNum;
  long       numLPrimes, Tsize, lpHashSize, hash0, hash1, newB;
  long       sF, LRpos, LApos, *d2;
  char       ifName[64], ofName[64], tmpName[64], byte;
  FILE      *ifp, *ofp;
  mmi_table_t *Lprimes, *T;
  rel_list   *RL;

  /* First, make the row files, with the actual primes instead of the indices. */
  printf("Doing makeRows()...\n");
  for (i=0; i<prelF->numFiles; i++) {
    sprintf(ofName, "%s.%d", rowF->prefix, i);
    RL = getRelList(prelF, i);
    if (!(ofp = fopen(ofName, "wb"))) {
      fprintf(stderr, "makeRows(): Could not open %s for write!\n", ofName);
      return;
    }
    for (l=0; l<RL->numRels; l++) {
      /* All we really need is the large primes; we should pull them
         directly out of the relData instead. But this will suffice for now.
      */
      sF = RL->relData[RL->relIndex[l]];
      LRpos = RL->relIndex[l] + 1 + 2*((int)GETNUMRFB(sF) + (int)GETNUMAFB(sF) + 
               (int)GETNUMSPB(sF) + 2);
      LApos = LRpos + GETNUMLRP(sF);

      for (j=0; j<GETNUMLRP(sF); j++) {
        if (RL->relData[LRpos+j] > 1) {
            data[j] = RL->relData[LRpos+j];
        } else data[j]=0;
      }
      for ( ; j<MAX_LARGE_RAT_PRIMES; j++)
        data[j]=0;
      d2 = data+MAX_LARGE_RAT_PRIMES;
      for (j=0; j<GETNUMLAP(sF); j++) {
        if (RL->relData[LApos+2*j] > 1) {
            d2[2*j] = RL->relData[LApos+2*j];
            d2[2*j+1] = RL->relData[LApos+2*j+1];
        } else {
          d2[2*j]=d2[2*j+1]=0;
        }
      }
      for ( ; j<MAX_LARGE_ALG_PRIMES; j++)
        d2[2*j]=d2[2*j+1]=0;
      byte = 0x00;
      writeRowLF(ofp, byte, data, dataSize);
    }
    fclose(ofp);
    clear_rel_list(RL);
    free(RL);
  }

  /*****************************************************************/
  /* Load up the Lprimes (leftover primes, that are not indexed by */
  /* their hash due to collision.                                  */
  /*****************************************************************/
  numLPrimes = loadLPrimes(&Lprimes, lpF);
  printf("makeRows() Read %ld primes from L file.\n", numLPrimes);

  /**********************************************************/
  /* Now, go through and replace the primes with indicies.  */
  /* Unforunately, we must do this in several passes, since */
  /* cannot presume to be able to fit all the large primes  */
  /* and indicies in memory at once.                        */
  /**********************************************************/
  Tsize = lpF->data1;
  lpHashSize  = lpF->data2;
  if (!(T = (mmi_table_t *)malloc(Tsize*sizeof(mmi_table_t)))) {
    fprintf(stderr, "makeRows() Error allocating %ldMB for T table!\n",
            Tsize*sizeof(mmi_table_t)/1048576);
    free(Lprimes); exit(-1);
  }
  for (i=0; i<lpF->numFiles; i++) {
    hash0 = i*lpF->data1;
    hash1 = (i+1)*lpF->data1;
    /* Load the T table. */
    loadTtable(T, Tsize, lpF, i);
    printf("Loaded T table: %s.%d to handle primes with: %ld <= hash < %ld.\n", 
            lpF->prefix, i, hash0, hash1);

    for (j=0; j<rowF->numFiles; j++) {
      sprintf(ifName, "%s.%ld", rowF->prefix, j);
      if (!(ifp = fopen(ifName, "rb"))) {
        fprintf(stderr, "makeRows() Error opening %s for read!\n", ifName);
        exit(-1);
      } 
      strcpy(tmpName, TMP_FILE);
      if (!(ofp = fopen(tmpName, "wb"))) {
        fprintf(stderr, "makeRows() Error opening temp file %s for write!\n", tmpName);
        exit(-1);
      }
      /****************************************************************/
      /* Now, read each row one at a time. 1st, check the byte field, */
      /* to see which fields are primes and which have already been   */
      /* converted to indicies. If a field still contains a prime,    */
      /* hash it, and see if the hash value is in the proper range.   */
      /****************************************************************/
      while (!(feof(ifp))) {
        if (readRowLF(&byte, data, dataSize, ifp)==0) {
          for (k=0; k<MAX_LARGE_RAT_PRIMES; k++) {
            if ((byte&BIT(k))==0) {
              /* This large prime has not yet been converted to an index.       */
              /* But first: Is it a large prime at all, or just a space-holder? */
              p = data[k];
              if (p>1) {
                h = NFS_HASH_Q(p, -1, lpHashSize);
                if ((h>=hash0) && (h<hash1)) {
                  /* It's in the proper range. Find it's index. */
                  if ((T[h-hash0].p == p) && (T[h-hash0].r==-1)) {
                    index = T[h-hash0].index;
                  } else {
                    index = findLPrime(Lprimes, numLPrimes, p, -1);
                    if (index < 0) {
                      fprintf(stderr, "makeRows() Error: Could not find rational prime %ld in master index!\n", p);
                      exit(-1);
                    }
                    else 
                      index = Lprimes[index].index;
                  }
                  data[k] = index;
                  byte ^= BIT(k); /* Signifies that we converted it to an index. */
                }
              }
            }
          }
          for (k=0; k<MAX_LARGE_ALG_PRIMES; k++) {
            if ((byte&BIT(MAX_LARGE_RAT_PRIMES+k))==0) {
              /* This large prime has not yet been converted to an index.       */
              /* But first: Is it a large prime at all, or just a space-holder? */
              p = data[MAX_LARGE_RAT_PRIMES + 2*k];
              r = data[MAX_LARGE_RAT_PRIMES + 2*k+1];
              if (p>1) {
                h = NFS_HASH_Q(p, r, lpHashSize);
                if ((h>=hash0) && (h<hash1)) {
                  /* It's in the proper range. Find it's index. */
                  if ((T[h-hash0].p == p) && (T[h-hash0].r==r)) {
                    index = T[h-hash0].index;
                  } else {
                    index = findLPrime(Lprimes, numLPrimes, p, r);
                    if (index < 0) {
                      fprintf(stderr, "makeRows() Error: Could not find alg. prime (%ld,%ld) in master index!\n", p,r);
                      exit(-1);
                    }
                    else 
                      index = Lprimes[index].index;
                  }
                  data[MAX_LARGE_RAT_PRIMES + 2*k] = index;
                  data[MAX_LARGE_RAT_PRIMES + 2*k+1] = 0;
                  byte ^= BIT(MAX_LARGE_RAT_PRIMES+k); /* Signifies that we converted it to an index. */
                }
              }
            }
          }
          /* Now, write it back out to the temp file. */
          writeRowLF(ofp, byte, data, dataSize);
        }
      }
      fclose(ifp); fclose(ofp);
      /* Now, replace the old file with the temp file. */
      remove(ifName); rename(tmpName, ifName);
    } /* end for (j=0... */
  }

  /*********************************************************************/
  /* Now, all the large primes have been replaced by their indicies.   */
  /* We should now go through and clean up the row files. That is,     */
  /* we won't need the byte field, and we don't need the fields that   */
  /* don't actually correspond to a large prime (i.e., right now,      */
  /* each entry has MAX_LARGE_RAT_PRIMES + MAX_LARGE_ALG_PRIMES fields */
  /* regardless of how many large primes the relation actually has.    */
  /***************************************************************************/
  /* The new format of this file will be:                                    */
  /* [# large primes][# relations][lPIndex_1]...[lPIndex_r][rel_1]...[rel_s] */
  /* Where the first 2 fields are single bytes and the remaining fields are  */
  /* longs. Note: The exact size of an entry is variable, depending on the   */
  /* first two fields.                                                       */
  /***************************************************************************/
  /* Note: We are assuming that, for now, row_i still corresponds to rel_i.  */
  /***************************************************************************/
  relNum = 0;
  initialFF=0;
  for (j=0; j<rowF->numFiles; j++) {
    sprintf(ifName, "%s.%ld", rowF->prefix, j);
    if (!(ifp = fopen(ifName, "rb"))) {
      fprintf(stderr, "makeRows() Error opening %s for read!\n", ifName);
      exit(-1);
    } 
    strcpy(tmpName, TMP_FILE);
    if (!(ofp = fopen(tmpName, "wb"))) {
      fprintf(stderr, "makeRows() Error opening temp file %s for write!\n", tmpName);
      exit(-1);
    }
    while (!(feof(ifp)) && (readRowLF(&byte, data, dataSize, ifp)==0)) {
      newB=0;
      for (k=0; k<MAX_LARGE_RAT_PRIMES; k++) {
        index = data[k];
        if (byte&BIT(k)) {
          newData[newB] = index;
          newB += 1;
        }
      }
      for (k=0; k<MAX_LARGE_ALG_PRIMES; k++) {
        index = data[MAX_LARGE_RAT_PRIMES + 2*k];
        if (byte&BIT((k+MAX_LARGE_RAT_PRIMES))) {
          newData[(int)newB] = index;
          newB += 1;
        }
      }
      newData[(int)newB] = relNum++;
      writeRowSF(ofp, newB, 1, newData);
      if (newB==0)
        initialFF++;
    }
    fclose(ifp); fclose(ofp);
    /* Now, replace the old file with the temp file. */
    remove(ifName); rename(tmpName, ifName);
  }
  printf("Row files have been written.\n");    

  free(Lprimes); free(T);
    
}

typedef struct {
  char numLP;
  char numRel;
  long lPIndex[MAX_LARGE_RAT_PRIMES+MAX_LARGE_ALG_PRIMES];
  long rels[MAX_RELS_IN_FF];
} sp_row_t;

/*********************************************************************/
int removeEvens(long *list, int size)
/*********************************************************************/
/* Given a list of integers: list[0], ..., list[size-1], reduce it   */
/* so it contains only those integers that appeared an odd number    */
/* of times. That is, remove the elements that occur an even number  */
/* of times.                                                         */
/* Return value: The number of elements in the reduced list.         */
/*********************************************************************/
/* This is a crappy way to do it, but it's good for now.             */
{ int j, k, unique;

  qsort(list, size, sizeof(long), cmpLongs);
  j=k=unique=0;
  while (j<size) {
    k=1;
    while (((j+k)<size) && (list[j] == list[j+k]))
      k++;
    if (k&0x01) {
      /* It occurrs an odd number of times, so keep it. */
      list[unique++] = list[j];
    }
    j += k;
  }
  return unique;
}

/******************************************************************************/
long doRowOps2(llist_t *P, llist_t *R, multi_file_t *rowF, long *numFF, long numCols)
/******************************************************************************/
/* The idea here is this:                                                     */
/* (1) Parition the columns into K subsets of manageable size.                */
/* (2) For each subset of the columns do as follows:                          */
/*     (2.0) Initialize an array of indicies, C, with |subset| entries to -1. */
/*     (2.1) Scan all row files for entries with exactly one large prime.     */
/*           If the large prime from such a row is in the current column      */
/*           range, do the following:                                         */
/*        (2.1.1) if C[pIndex - offset] >= 0, do nothing.                     */
/*        (2.1.2) if C[pIndex - offset] == -1, store this row in RAM and index*/
/*                its location : C[pIndex-offset] <-- loc >= 0.               */
/*     (2.2) Scan all row files again. For each row read, do the following:   */
/*        (2.2.1) Check if it has any large primes with an index in the       */
/*           current column range. If so, check the C[index-offset] to see if */
/*           there is a single-large prime row with the matching prime. If    */
/*           there is one, and it's different than the row we're looking at,  */
/*           use it to remove that prime from this row.                       */
/*        (2.2.2) Append the current row to a temp file (either the row is    */
/*           unchanged, or we tweaked it to remove a large prime).            */
/*        (2.2.3) After all rows from the file have been read and checked,    */
/*           replace the original file with the temp file. Wipe the temp file */
/*           and move on to the next row file.                                */
/* (3) If any reductions were made, return 1. Otherwise, return 0.            */
/******************************************************************************/
{ long  j, data[256], numFull, ff=0, b1, b2;
  char  fName[MAXFNAMESIZE+1];
  FILE *ifp;


  ll_init(P, 1024*1024, 1024*1024/4);

  for (j=0; j<rowF->numFiles; j++) {
    sprintf(fName, "%s.%ld", rowF->prefix, j);
    if (!(ifp = fopen(fName, "rb"))) {
      fprintf(stderr, "doRowOps() Error opening %s for read!\n", fName);
      exit(-1);
    }
    printf("Reading large primes from relations in file %ld/%d.\n", j+1, rowF->numFiles);
    while (!(feof(ifp))) {
      if (readRowSF(&b1, &b2, data, ifp)==0) {
        /* Get its large primes. */
        if (b1==0) ff++;
        ll_appendField(P, data, b1);
      }
    }
    printf("Done.\n");
    fclose(ifp);
  }   
  { long f=0, i;
    for (i=0; i<P->numFields; i++) 
      if ((P->index[i+1]-P->index[i])==0) f++;
    printf("Before combParts(), there are %ld full relations (ff=%ld).\n", f, ff);
  }
  numFull = combParts(R, P);

  return numFull;
}

/*********************************************************************/
long getCols(char *colName, multi_file_t *prelF, multi_file_t *lpF, nf_t *N, long totalRels,
             long minFull)
/*********************************************************************/
{ multi_file_t rowF;
  long         i, j, k, l, numLargeP, thisSize, R0, R1;
  long         numFulls, rIndex, numFF;
  long         aOffset, spOffset, t;
  char         fName[64], prelName[64], indexName[64];
  FILE         *fp, *ofp, *ofp2;
  column_t     C;
  rel_list     RL;
  relation_t   R;
  struct stat fileInfo;
  mpz_t       tmp;
  llist_t     P, Rl;

  mpz_init(tmp);
  rowF.numFiles = prelF->numFiles;
  strcpy(rowF.prefix, "rows");
  
  numLargeP = makeMasterIndex2(lpF, prelF, totalRels);
  makeRows(&rowF, prelF, lpF);

  printf("***************************\n");
  printf("There are %ld large primes versus %ld relations.\n", 
          numLargeP, initialRelations);
  msgLog(NULL, "largePrimes: %ld , relations: %ld", numLargeP, initialRelations);
  printf("***************************\n");
  if ((numLargeP > 1.2*initialRelations || (numLargeP - initialRelations  > 2.0*minFull )) 
      && (!(forceCC))) {
    /* All done. We have no more use for the row files. */
    for (i=0; i<rowF.numFiles; i++)  {
      sprintf(fName, "%s.%ld", rowF.prefix, i);
      remove(fName);
    }
    return 0;
  }

  numFulls = doRowOps2(&P, &Rl, &rowF, &numFulls, numLargeP);

  if (numFulls < minFull) {
    for (i=0; i<rowF.numFiles; i++) {
      sprintf(fName, "%s.%ld", rowF.prefix, i);
      remove(fName);
    }
    ll_clear(&P); ll_clear(&Rl);
    return numFulls;
  }

  /**************************************************************/
  /* Convert the full relations into column_t data.             */
  /* This is done in several passes. First, create the column   */
  /* file, with only the relations. Later, we will go through   */
  /* and replace the relations with their corresponding primes. */
  /**************************************************************/
  numFF=0;
  C.numPrimes = 0x00; C.QCB[0] = C.QCB[1] = 0x00000000;
  sprintf(indexName, "%s.index", colName);
  if (!(ofp = fopen(colName, "wb"))) {
    fprintf(stderr, "getCols() Error opening %s for write!\n", colName);
    exit(-1);
  }
  if (!(ofp2 = fopen(indexName, "wb"))) {
    fprintf(stderr, "getCols() Error opening %s for write!\n", indexName);
    fclose(ofp);
    exit(-1);
  }

  for (i=0; i<P.numFields; i++) {
    if ((Rl.index[i+1]-Rl.index[i]>0) && (P.index[i+1]==P.index[i])) {
      /* This is a full relation, coming from several (a,b) pairs. */
      C.numRels = (int)(Rl.index[i+1] - Rl.index[i]);
      for (j=0; j<C.numRels; j++)
        C.Rels[j] = Rl.data[Rl.index[i]+j];
      writeColLF(ofp, &C);
      writeColIndex(ofp2, &C);
      numFF++;
    }
  }

  fclose(ofp); fclose(ofp2);
  printf("After re-scanning files and building column indicies, numFF=%ld.\n", numFF);
  /*******************************************************************************/
  /* We now have a completely unprocessed column file. It has only the relations */
  /* occurring in each column. We need to go through and replace each relation   */
  /* with the corresponding primes, QCB bits.                                    */
  /*******************************************************************************/

  /* First, see how much room we'll need for the largest */
  /* of the processed relation files.                    */
  RL.maxDataSize = 0;
  RL.maxRels = 1.1*((double)totalRels/(double)prelF->numFiles);
  for (i=0; i<prelF->numFiles; i++) {
    sprintf(prelName, "%s.%ld", prelF->prefix, i);
    if (stat(prelName, &fileInfo)) {
      printf("Could not stat file %s!\n", prelName);
      exit(-1);
    } else {
      thisSize = fileInfo.st_size;
      RL.maxDataSize = MAX(thisSize/sizeof(long), RL.maxDataSize);
      t=0;
      if ((fp = fopen(prelName, "rb"))) {
        readRawLong(&t, fp);
        fclose(fp);
      }
      RL.maxRels = MAX(RL.maxRels, t);
    }
  }
  RL.maxRels += 1;
  /* Now allocate for the relations. */
  if (!(RL.relData = (long *)malloc(RL.maxDataSize * sizeof(long)))) {
    fprintf(stderr, "Error allocating %ldMB for re-reading relations!\n",
            RL.maxDataSize * sizeof(long)/1048576);
    exit(-1);
  }
  printf("Allocated %ldMB for re-reading relations.\n", RL.maxDataSize*sizeof(long)/1048576);
  if (!(RL.relIndex = (long *)malloc(RL.maxRels * sizeof(long)))) {
    fprintf(stderr, "Error allocating %ldMB for relation pointers!\n", 
            RL.maxRels * sizeof(long)/1048576);
    free(RL.relData);
    exit(-1);
  }
  aOffset = N->FB->rfb_size;
  spOffset = aOffset + N->FB->afb_size;


  printf("Creating %ld matrix columns...\n", numFF);
  strcpy(fName, TMP_FILE);
  R0=R1=0;
  for (i=0; i<prelF->numFiles; i++) {
    sprintf(prelName, "%s.%ld", prelF->prefix, i);
    RL.numRels = 0;
    readRelList(&RL, prelName);
    R1 = R0 + RL.numRels;
    printf("Re-read %ld relations from %s :  [%ld, %ld).\n", RL.numRels, prelName, R0, R1);
    /* Now, we have in RAM the relations numbered [R0, R1). */
    if (!(fp = fopen(colName, "rb"))) {
      fprintf(stderr, "getCols() Error opening %s for read!\n", colName);
      exit(-1);
    }
    if (!(ofp = fopen(fName, "wb"))) {
      fprintf(stderr, "getCols() Error opening %s for write!\n", fName);
      fclose(fp);
      exit(-1);
    }

    for (j=0; j<numFF; j++) {
      if (readColLF(&C, fp)==0) {
        for (k=0; k<C.numRels; k++) {
          if ((C.Rels[k]>= R0) && (C.Rels[k] < R1)) {
            /* We have the factorization of this relation in RAM. */
            rIndex = C.Rels[k] - R0;
            for (l=k; l<C.numRels; l++)
              C.Rels[l] = C.Rels[l+1];
            C.numRels -= 1;
            /****************************************************************/        
            /* This is horribly inefficient. If we get ambitious, we should */
            /* replace this with code that pulls the factors directly out   */
            /* of the data. But for now, it'll do.                          */
            /****************************************************************/        
            dataConvertToRel(&R, &RL.relData[RL.relIndex[rIndex]]);
            /* Note: These '%2's used to be `&0x01', but negative exponents can occur!   */
            /* In particular, we cannot guarantee that the internal representation makes */
            /* the small bit 1 for negative odd numbers!                                 */
            if (C.numPrimes + R.rFSize + R.aFSize + R.spSize >= MAX_ROWS_IN_COL) {
              fprintf(stderr, "getCols(): MAX_ROWS_IN_COL exceeded!\n");
              exit(-1);
            }
            for (l=0; l<R.rFSize; l++) 
              if (R.rExps[l]%2) {
                C.rows[C.numPrimes] = R.rFactors[l]; C.numPrimes += 1;
              }
            for (l=0; l<R.aFSize; l++) 
              if (R.aExps[l]%2) {
                C.rows[C.numPrimes] = aOffset+ R.aFactors[l]; C.numPrimes += 1;
              }
            for (l=0; l<R.spSize; l++) 
              if (R.spExps[l]%2) {
                C.rows[C.numPrimes] = spOffset+ R.spFactors[l]; C.numPrimes += 1;
              }
            C.QCB[0] ^= R.qcbBits[0]; C.QCB[1] ^= R.qcbBits[1];
            C.numPrimes = removeEvens(C.rows, C.numPrimes);
            k--; /* Since we removed this relation and shifted everything down,
                    we need this so that after the increment, we look at the
                    same position (which will be the next relation index, if there
                    is another.)
                 */
          }
        }
        writeColLF(ofp, &C);
      }
    }
    R0 += RL.numRels;
    fclose(fp); fclose(ofp);
    remove(colName); rename(fName, colName);
  }
  free(RL.relData); free(RL.relIndex);

  /* Finally, go back through and clean up the column file - now that none */
  /* of the entries have any un-resolved relations left, all we need is    */
  /* the prime-indexing fields.                                            */
  if (!(fp = fopen(colName, "rb"))) {
    fprintf(stderr, "getCols() Error opening %s for read!\n", colName);
    exit(-1);
  }
  strcpy(fName, TMP_FILE);
  if (!(ofp = fopen(fName, "wb"))) {
    fprintf(stderr, "getCols() Error opening %s for write!\n", TMP_FILE);
    fclose(fp);
    exit(-1);
  }
  fwrite(&numFF, sizeof(long), 1, ofp);
  for (j=0; j<numFF; j++) {
    /* This might be where the MingW32 version is going astray.
       Something seems to be getting screwed in the translation
       from LF format to SF format.
    */
    readColLF(&C, fp);
    writeColSF(ofp, &C);
  }
  fclose(fp); fclose(ofp);
  remove(colName); rename(fName, colName);
  for (i=0; i<rowF.numFiles; i++)  {
    sprintf(fName, "%s.%ld", rowF.prefix, i);
    remove(fName);
  }


  ll_clear(&Rl);
  ll_clear(&P);
  mpz_clear(tmp);
  return numFF;
}

/******************************************************/
int set_prelF(multi_file_t *prelF, long maxFileSize)
/******************************************************/
/* Check to see how many files there are, and whether */
/* or not this needs to be increased. If so, increase.*/
/******************************************************/
{ int    i, j, cont, newFiles;
  struct stat fileInfo;
  char   fName[512], newName[512];
  long   maxSize=0, where, newWhere;
  long   k;
  rel_list   RL, newRL;
  relation_t R;

  i=0;
  do {
    cont=0;
    sprintf(fName, "%s.%d", prelF->prefix, i);
    printf("Checking file %s ...\n", fName);
    if (stat(fName, &fileInfo)==0) {
      maxSize = MAX(maxSize, fileInfo.st_size);
      cont=1;
      i++;
    }
  } while (cont);
  prelF->numFiles = MAX(i, 1);
  printf("Largest prel file size is %ld versus max allowed of %ld.\n", 
          maxSize, maxFileSize);
  if (maxSize > maxFileSize) {
    /* We need to increase the number of files, which means
       moving the existing data around.
    */
    newFiles = prelF->numFiles+2;
    printf("Increasing prelF->numFiles from %d to %d...\n",
            prelF->numFiles, newFiles);
    RL.maxDataSize = 1000 + maxSize/sizeof(long);
    if (!(RL.relData = (long *)malloc(RL.maxDataSize * sizeof(long)))) {
      fprintf(stderr, "Error allocating %ldMB for processed relation file!\n",
              RL.maxDataSize * sizeof(long)/1048576);
      exit(-1);
    }
    RL.maxRels = (unsigned long)RL.maxDataSize/20;
    if (!(RL.relIndex = (long *)malloc(RL.maxRels * sizeof(long)))) {
      fprintf(stderr, "Error allocating %ldMB for relation pointers!\n",
              RL.maxRels * sizeof(long)/1048756);
      exit(-1);
    }
    newRL.maxDataSize = maxSize/sizeof(long);
    if (!(newRL.relData = (long *)malloc(newRL.maxDataSize * sizeof(long)))) {
      fprintf(stderr, "Error allocating %ldMB for processed relation file!\n",
              newRL.maxDataSize * sizeof(long)/1048576);
      exit(-1);
    }
    newRL.maxRels = (unsigned long)newRL.maxDataSize/20;
    if (!(newRL.relIndex = (long *)malloc(newRL.maxRels * sizeof(long)))) {
      fprintf(stderr, "Error allocating %ldMB for relation pointers!\n",
              newRL.maxRels * sizeof(long)/1048756);
      exit(-1);
    }
    for (i=0; i<prelF->numFiles; i++) {
      sprintf(fName, "%s.%d", prelF->prefix, i);
      readRelList(&RL, fName);
      printf("Read %ld relations from %s.\n", RL.numRels, fName);
      for (j=0; j<newFiles; j++) {
        sprintf(fName, ".%s.%d", prelF->prefix, j);
        readRelList(&newRL, fName); 
        where=0; newWhere = newRL.relIndex[newRL.numRels];

        for (k=0; k<RL.numRels; k++) {
          where += dataConvertToRel(&R, &RL.relData[where]);
          if (NFS_HASH(0,R.b,newFiles)==j) {   
            newWhere += relConvertToData(&newRL.relData[newWhere], &R);
            newRL.numRels += 1;
            newRL.relIndex[newRL.numRels] = newWhere;
          }
        }
        /* Write the new file: */
        writeRelList(fName, &newRL);
      }
      /* Delete the old i-th file. */
      sprintf(fName, "%s.%d", prelF->prefix, i);
      remove(fName);
    }
    prelF->numFiles = newFiles;
    for (i=0; i<prelF->numFiles; i++) {
      /* rename the files. */
      sprintf(fName, ".%s.%d", prelF->prefix, i);
      sprintf(newName, "%s.%d", prelF->prefix, i);
      rename(fName, newName);
    }
    free(RL.relData); free(RL.relIndex);
    free(newRL.relData); free(newRL.relIndex);
  }
  return 0;
}

/*
  This function should be tweaked to read the input file in blocks
  of some maximum size at a time, instead of the whole file at once!
  I believe this is the cause of the memory problems some folks are having.
*/

/*************************************************************/
long addNewRelations4(rel_list *RL, char *fName, long hashVal, long hashSize, nf_t *N)
/*************************************************************/
/* Read (a,b) pairs from fName, and append them to RL        */
/* if they have the proper hash value.                       */
/* Return value is the number of relations read and          */
/* factored.                                                 */
/*************************************************************/
{ long       numNew=0, where, sF, numRead=0;
  relation_t R;
  FILE      *fp;
  int        factRes, i;
  char       thisLine[512];
  double     startTime, now;
  long       nextReportNumRead = 10000;
  long       fSize, fBlockSize, fRemainSize, fTotalRead=0;
  unsigned char *fData, *fPos, *fLimit = NULL, *fEol = NULL;
  unsigned char *fWarningTrack=NULL;
  static char xdigit[256] = {
    -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x00+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x10+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x20+ */
     0,  1,  2,  3,  4,  5,  6,  7,  8,  9, -1, -1, -1, -1, -1, -1, /* 0x30+ */
    -1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x40+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x50+ */
    -1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x60+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x70+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x80+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0x90+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0xa0+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0xb0+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0xc0+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0xd0+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0xe0+ */
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* 0xf0+ */
  };
  nfs_fb_t *FB = N->FB;
  long maxRFB = FB->rfb[2 * (FB->rfb_size - 1)];
  long maxAFB = FB->afb[2 * (FB->afb_size - 1)];

  if (!(fp = fopen(fName, "r"))) {
    fprintf(stderr, "Error opening %s for read.\n", fName);
    return 0;
  }
  /* First, find where the data should go: */
  where = 0;
  if (RL->numRels > 0) {
    where = RL->relIndex[RL->numRels-1];
    sF = RL->relData[where];
    where += LONGS_IN_ENTRY(sF);
  } 
  RL->relIndex[RL->numRels] = where;
  fseek(fp, 0, SEEK_END);
  fSize = ftell(fp);
  if (fSize == 0) {
    fclose(fp);
    return 0;
  }
  fseek(fp, 0, SEEK_SET);
  fBlockSize = MIN(fSize, MAX_SPAIRS_ALLOC);
  if ((fData = (unsigned char *)malloc(fBlockSize + 1)) != NULL) {
    fTotalRead = fread(fData, 1, fBlockSize, fp);
    fLimit = fWarningTrack = fData + fBlockSize;
    if (fBlockSize >= MAX_SPAIRS_ALLOC)
      fWarningTrack -= 512; /* Where we stop and read the next block from file. */
    *fLimit = '\0';
    fEol = fData - 1;
  }
  startTime = sTime();
  for (;;) {
    long p, r, k;
    int c, m;


    if (fData != NULL) {
      fPos = fEol + 1;
      if (fPos >= fLimit) {
        break;
      } else if ((fPos > fWarningTrack)&&(fTotalRead < fSize)) {
        /* Read the next block of data from file. */
        fRemainSize = fLimit-fPos;
        memmove(fData, fPos, fRemainSize*sizeof(char));
        fBlockSize = fRemainSize + fread(fData+fRemainSize, 1, fBlockSize-fRemainSize,fp);
        fTotalRead += fBlockSize;
        fPos = fData;
//        fEol = fData - 1;
        fLimit = fData + fBlockSize;
        *fLimit = '\0';
      }
    } else {
      if (feof(fp)) {
        break;
      }
      thisLine[0] = '\0';
      fgets(thisLine, 512, fp);
      fPos = thisLine;
    }
    for (fEol = fPos; *fEol && *fEol != '\n'; fEol++) { /* search end-of-line */
      ;
    }
    if (*fPos == '#') { /* comment line */
      continue;
    }
    *fEol = '\0';
    /* expand and optimize parseOutputLine() */
    /* %ld */
    fPos += (m = *fPos == '-');
    if ((unsigned int)(p = *fPos - '0') >= 10) {
      continue;
    }
    fPos++;
    while ((unsigned int)(c = *fPos - '0') < 10) {
      p = (((p << 2) + p) << 1) + c; /* p = p * 10 + c */
      fPos++;
    }
    R.a = m ? -p : p;
    /* , */
    fPos++;
    /* %ld */
    fPos += (m = *fPos == '-');
    if ((unsigned int)(p = *fPos - '0') >= 10) {
      continue;
    }
    fPos++;
    while ((unsigned int)(c = *fPos - '0') < 10) {
      p = (((p << 2) + p) << 1) + c; /* p = p * 10 + c */
      fPos++;
    }
    R.b = m ? -p : p;
    /* : */
    while (fPos < fEol && *fPos++ != ':') {
      ;
    }
    /* %lx,%lx,... */
    R.p[0] = R.p[1] = 1;
    R.rFSize = 0;
    m = 0;
    while (fPos < fEol && *fPos != ':') {
      if ((p = xdigit[*fPos++]) < 0) {
        continue;
      }
      while ((c = xdigit[*fPos]) >= 0) {
        p = (p << 4) + c;
        fPos++;
      }
      k = lookupRFB(p, FB);
      if (k >= 0) {
        R.rFactors[R.rFSize++] = k;
      } else if ((p > maxRFB) && (m < FB->maxLP)) {
        R.p[m++] = p;
      }
    }
    /* : */
    fPos += *fPos == ':';
    /* %lx,%lx,... */
    R.a_p[0] = R.a_p[1] = 1;
      R.a_r[0] = R.a_r[1] = 1;
    R.aFSize = 0;
    m = 0;
    while (fPos < fEol && *fPos != ':') {
      if ((p = xdigit[*fPos++]) < 0) {
        continue;
      }
      while ((c = xdigit[*fPos]) >= 0) {
        p = (p << 4) + c;
        fPos++;
      }
      if (R.b % p) { /* p is always non-zero? */
        r = mulmod32(p + (R.a % p), inverseModP(R.b, p), p);
        k = lookupAFB(p, r, FB);
        if (k >= 0) {
          R.aFactors[R.aFSize++] = k;
        } else if ((p > maxAFB) && (m < FB->maxLPA)) {
          R.a_p[m] = p; 
          R.a_r[m++] = r;
        }
      }
    }
    if (NFS_HASH(0, R.b, hashSize) != hashVal) {
      continue;
    }
    numRead++;
    factRes = completePartialRelFact(&R, N, CLIENT_SKIP_R_PRIMES, CLIENT_SKIP_A_PRIMES);
    if (factRes == 0) {
      where += relConvertToData(&RL->relData[where], &R);
      RL->numRels++;
      RL->relIndex[RL->numRels] = where;
      numNew++;
    } else {
      printf("Relation (%ld, %ld) bad : return value %d.\n", R.a, R.b, factRes);
    }
    /* avoid calling clock() over again */
    if (numRead >= nextReportNumRead) {
      nextReportNumRead += 10000;
      now = sTime();
      i = printf("Status: processed %ld relations from %s... (at %1.2lf rels/sec)",
                 numRead, fName,
                 /* considering now == startTime */
                 now != startTime ? (double)numRead / (now - startTime) : 0.0);
      fflush(stdout);
      while (i--) {
        printf("\b");
      }
    }
  }
  printf("\n");
fclose(fp);
  if (fData != NULL) {
    free(fData);
  } else {
//    fclose(fp);
  }
  return numNew;
}


/**************************************************************************/
long makeMasterIndex2(multi_file_t *lpF, multi_file_t *relF, long estN)
/**************************************************************************/
/* lpFPrefix is the filename prefix where the indices are to be stored.   */
/* relPRefix is the filename prefix for the `numRelFiles' relation files. */
/* estN is an estimate for the number of large primes. For this estimate, */
/* it's better to err on the high side than the low side.                 */
/* Return value: The number of large primes found, or -1 on error.        */
/**************************************************************************/
/* It's assumed that the filenames are reasonably sized (< 60 bytes).     */
/**************************************************************************/
{ long         i, j, k=0, K, Tsize, h, N=0, p, r, entry, numRelsInFile;
  long         _p, _r, _h, _ct, _index, hashSize;
  mmi_table_t *T;
  char         fName[64], LfName[64];
  FILE        *fp;
  int          res=0, alg;
  relation_t   R;

  hashSize = getNextPrime(4*estN);
  K = sizeof(mmi_table_t)*hashSize/MAX_MEM_ALLOC + 1;
  lpF->numFiles = K;
  Tsize = (hashSize/K);
  lpF->data1 = Tsize;
  lpF->data2 = hashSize;
  if (!(T = (mmi_table_t *)malloc(Tsize*sizeof(mmi_table_t)))) {
    fprintf(stderr, "makeMasterIndex() Error allocating %ldMB for the T table!\n",
            Tsize*sizeof(mmi_table_t)/1048576);
    exit(-1);
  }
  sprintf(LfName, "%s.L", lpF->prefix);
  if (!(fp = fopen(LfName, "wb"))) {
    fprintf(stderr, "makeMasterIndex(): Error opening file %s for write!\n", LfName);
    res = -1;
    goto MMI_SS_DONE;
  }
  fclose(fp);

  N=0;
  for (j=0; j<K; j++) {
    printf("Large prime pass %ld of %ld.\n", j+1, K);
    /* Initialize T: */
    for (i=0; i<Tsize; i++) {
      T[i].p = 0; T[i].r = -1;
      T[i].index = -1; T[i].counter = 0;
    }

    for (i=0; i<relF->numFiles; i++) {
      sprintf(fName, "%s.%ld", relF->prefix, i);
      if (!(fp = fopen(fName, "rb"))) {
        fprintf(stderr, "makeMasterIndex() Error opening relation file %s for read! Ignoring.\n", fName);
      } else {
        long fSize, *fData;
        fseek(fp, 0, SEEK_END);
        fSize = ftell(fp) / sizeof(long);
        fseek(fp, 0, SEEK_SET);
        if (fSize > 0 && (fData = (long *)malloc(sizeof(long) * fSize)) != NULL) {
          long *fPos = fData, *fLimit = fData + (fSize - 1);
#define G_SIZE (4 * 10 * 1024)
          long *gData = (long *)malloc(sizeof(long) * G_SIZE);
          long *gPos = gData, *gLimit = gData + G_SIZE;
          FILE *gp = fopen(LfName, "ab");
          printf("Reading whole relation file %s\n", fName);
          fread(fData, sizeof(long), fSize, fp);
          fclose(fp);
          numRelsInFile = *fPos++;
          printf("Scanning %ld relations on memory...\n", numRelsInFile);
          while (fPos < fLimit) {
            long sF = *fPos++;
            int numLRP, numLAP;
            fPos += 2*(GETNUMRFB(sF) + GETNUMAFB(sF) + GETNUMSPB(sF) + 2);
            numLRP = (int)GETNUMLRP(sF);
            for (k = 0; k < numLRP; k++) {
              R.p[k] = *fPos++;
            }
            for (; k < MAX_LARGE_RAT_PRIMES; k++) {
              R.p[k] = 1;
            }
            numLAP = (int)GETNUMLAP(sF);
            for (k = 0; k < numLAP; k++)  {
              R.a_p[k] = *fPos++;
              R.a_r[k] = *fPos++;
            }
            for (; k < MAX_LARGE_ALG_PRIMES; k++) {
              R.a_p[k] = 1;
              R.a_r[k] = 0;
            }
            for (k = 0; k < MAX_LARGE_RAT_PRIMES && (p = R.p[k]) > 1; k++) {
              entry = NFS_HASH_Q(p, -1, hashSize) - j * Tsize;
              if (entry >= 0 && entry < Tsize) {
                if (T[entry].p == p && T[entry].r == -1) {
                  T[entry].counter++;
                } else if (T[entry].p == 0) {
                  T[entry].p = p;
                  T[entry].r = -1;
                  T[entry].counter = 1;
                } else {
                  *gPos++ = p;
                  *gPos++ = -1;
                  *gPos++ = 0;
                  *gPos++ = 0;
                  if (gPos >= gLimit) {
                    fwrite(gData, sizeof(long), G_SIZE, gp);
                    gPos = gData;
                  }
                }
              }
            }
            for (k = 0; k < MAX_LARGE_ALG_PRIMES && (p = R.a_p[k]) > 1; k++) {
              r = R.a_r[k];
              entry = NFS_HASH_Q(p, r, hashSize) - j * Tsize;
              if (entry >= 0 && entry < Tsize) {
                if (T[entry].p == p && T[entry].r == r) {
                  T[entry].counter++;
                } else if (T[entry].p == 0) {
                  T[entry].p = p;
                  T[entry].r = r;
                  T[entry].counter = 1;
                } else {
                  *gPos++ = p;
                  *gPos++ = r;
                  *gPos++ = 0;
                  *gPos++ = 0;
                  if (gPos >= gLimit) {
                    fwrite(gData, sizeof(long), G_SIZE, gp);
                    gPos = gData;
                  }
                }
              }
            }
          }
          if (gPos > gData) {
            fwrite(gData, sizeof(long), gPos - gData, gp);
          }
          fclose(gp);
          free(gData);
          free(fData);
        } else {
        readRawLong(&numRelsInFile, fp);
        printf("Scanning %ld relations from %s...\n", numRelsInFile, fName);
        while (readRel(&R, fp)==0) {
          k=0; alg=0;
          while ((alg==0) || ((k < MAX_LARGE_ALG_PRIMES) && (R.a_p[k] > 1))) {
            /* Get the next large prime in this relation, be it rational or algebraic. */
            p=0; r=-1;
            if ((alg==0) && (k < MAX_LARGE_RAT_PRIMES) && (R.p[k]>1)) {
              p = R.p[k]; r=-1;
            } else if (alg==0) {
              alg=1; k=0;
            }
            if (alg && (k < MAX_LARGE_ALG_PRIMES) && (R.a_p[k] > 1)) {
              p = R.a_p[k]; r = R.a_r[k];
            }
            k++;
            if (p > 1) {
              /* OK - we found a large prime. First, compute its hash value. */
              h = NFS_HASH_Q(p, r, hashSize);
              
              /* Is the hash value in the proper range for this output file? */
              if ((h >= j*Tsize) && (h < (j+1)*Tsize)) {

                /* It is: retrieve the corresponding entry. */
                entry = h - j*Tsize;
                if ((T[entry].p == p) && (T[entry].r == r))
                  T[entry].counter += 1;
                else if (T[entry].p == 0) {
                  T[entry].p = p; T[entry].r = r; T[entry].counter = 1;
                } else {
                  /* This prime is not in the hash table, but something else is in       */
                  /* it's place. Just append this prime to the L file for later sorting. */
                  appendToLFile(LfName, p, r);
                }
              }
            }
          }
        }
        fclose(fp); /* All done with this relation file. */
        }
      }
    }
    /* Set the index fields: */
    for (i=0; i<Tsize; i++) {
      if (T[i].p > 1) {
        T[i].index = N++;
      }
    }
    /* Finally, we should write the 'T' table out to file. */
    sprintf(fName, "%s.%ld", lpF->prefix, j);
    if (!(fp = fopen(fName, "wb"))) {
      fprintf(stderr, "makeMasterIndex(): Error opening file %s for write!\n", fName);
      res = -1;
      goto MMI_SS_DONE;
    }
    {
#define H_SIZE (5 * 10 * 1024)
      long *hData = (long *)malloc(sizeof(long) * H_SIZE);
      if (hData != NULL) {
        long *hPos = hData, *hLimit = hData + H_SIZE;
        for (i = 0; i < Tsize; i++) {
          if (T[i].p > 1) {
            *hPos++ = i + (j * Tsize);
            *hPos++ = T[i].p;
            *hPos++ = T[i].r;
            *hPos++ = T[i].index;
            *hPos++ = T[i].counter;
            if (hPos >= hLimit) {
              fwrite(hData, sizeof(long), H_SIZE, fp);
              hPos = hData;
            }
          }
        }
        if (hPos > hData) {
          fwrite(hData, sizeof(long), hPos - hData, fp);
        }
        free(hData);
      } else {
    for (i=0; i<Tsize; i++) {
      if (T[i].p > 1) {
        /* Perhaps we might also want to restrict our attention to primes
           appearing more than once?
        */
        _p = T[i].p; _r = T[i].r; _h = i+(j*Tsize);
        _index = T[i].index; _ct = T[i].counter;
        writeRawLong(fp, &_h); writeRawLong(fp, &_p);
        writeRawLong(fp, &_r); writeRawLong(fp, &_index);
        writeRawLong(fp, &_ct);
      }
    }
      }
    }
    fclose(fp);
    
  }
MMI_SS_DONE:
  printf("makeMasterIndex(): Found %ld large primes + those in %s.\n", N, LfName);
  printf("Tsize = %ld\n", Tsize);
  /* Now, fix the L table. That is, sort it, remove duplicates (and keep count), */
  /* and give the elements consistent indices.                                   */
  if (!(fp = fopen(LfName, "rb"))) {
    printf("Warning: Could not open %s for read.\n", LfName);
  } else {
    i=0;
    while ((i<Tsize) && (!(feof(fp)))) {
      _p = 1;
      readRawLong(&_p, fp); 
      readRawLong(&_r, fp); 
      readRawLong(&_index, fp);
      readRawLong(&_ct, fp);
      if (_p>1) {
        T[i].p = _p; T[i].r = _r; T[i].counter = 1;
        i++;
      }   
    }
    fclose(fp);
    printf("Read %ld entries from %s, counting repetitions.\n", i, LfName);
    if (i >= Tsize) 
      printf("Warning: read Tsize = %ld large primes from %s - there may have been more!\n", Tsize, LfName);
    if (i<=1) {
      k=i; goto SS_UP_DONE;
    }
    qsort(T, i, sizeof(mmi_table_t), cmpMMI_t);
    /* Now, reduce to unique, counting occurrences. */
    j=1; k=0;
    while (j<i) {
      if ((T[j].p==T[k].p) && (T[j].r == T[k].r)) {
        T[k].counter += 1;
        j++;
      } else {
        k++;
        T[k].p = T[j].p; T[k].r = T[j].r; T[k].counter = 1;
        j++;
      }
    }
    k++;
SS_UP_DONE:
    /* Now, the first k entries: T[0],...,T[k-1] are unique. */
    if ((fp = fopen(LfName, "wb"))) {
      for (i=0; i<k; i++) {
        _p = T[i].p; _r = T[i].r; _index = N+i; _ct = T[i].counter;
        writeRawLong(fp, &_p); 
        writeRawLong(fp, &_r); 
        writeRawLong(fp, &_index);
        writeRawLong(fp, &_ct);
      }
      fclose(fp);
    }
  }
  N += k;
  printf("Found %ld large primes in total.\n", N);

  free(T);
  if (res)
    return res;
  return N;
}


/****************************************************/
int main(int argC, char *args[])
/****************************************************/
{ char       fbName[64], prelName[40], newRelName[64], depName[64], colName[64];
  char       tmpStr[1024], str[128];
  int        i, qcbSize = DEFAULT_QCB_SIZE, seed=DEFAULT_SEED, retVal=0;
  double     startTime, stopTime, rStart, rStop;
  long       oldSize, newSize, maxSize, totalRels, numNewRels, bytesUsed;
  long       minFF;
  struct stat fileInfo;
  rel_list   RL;
  nf_t       N;
  mpz_fact_t D;
  rel_t     *Rp=NULL;
  multi_file_t prelF, lpF;
  FILE            *fp;

  prelF.numFiles=DEFAULT_NUM_FILES;
  fbName[0] = prelF.prefix[0] = newRelName[0] = 0;
  strcpy(depName, DEFAULT_DEPNAME);
  strcpy(colName, DEFAULT_COLNAME);
  strcpy(lpF.prefix, DEFAULT_LPI_NAME);

  printf(START_MSG, GGNFS_VERSION);

  for (i=1; i<argC; i++) {
    if (strcmp(args[i], "-fb")==0) {
      if ((++i) < argC) 
        strncpy(fbName, args[i], 64);
    } else if (strcmp(args[i], "-prel")==0) {
      if ((++i) < argC) 
        strncpy(prelF.prefix, args[i], 32);
    } else if (strcmp(args[i], "-newrel")==0) {
      if ((++i) < argC) 
        strncpy(newRelName, args[i], 64);
    } else if (strcmp(args[i], "-qs")==0) {
      if ((++i) < argC)
        qcbSize = atoi(args[i]);
    } else if (strcmp(args[i], "-seed")==0) {
      if ((++i) < argC) {
        seed = atoi(args[i]);
      }
    } else if (strcmp(args[i], "-nodfactor")==0) {
      discFact=0;
    } else if (strcmp(args[i], "-forceCC")==0) {
      forceCC=1;
    } else if (strcmp(args[i], "-v")==0) {
      verbose++;
    }
  }
  srand(seed);
  startTime = sTime();
  
  if ((fbName[0]==0) || (prelF.prefix[0]==0)) {
    printf("USAGE: %s %s\n", args[0], USAGE);
    exit(0);
  }
  msgLog("", "GGNFS-%s : getdeps", GGNFS_VERSION);
  sprintf(prelName, "%s.%d", prelF.prefix, 0);
  if (stat(prelName, &fileInfo)==0) {
    printf("It appears this is not the first run. Setting discFact=-1.\n");
    discFact=0;
  }


  initNF(&N);
  N.FB = (nfs_fb_t *)malloc(sizeof(nfs_fb_t));
  initFB(N.FB);
  mpz_fact_init(&D);
  if (loadFB(fbName, N.FB)) {
    printf("Could not load FB from %s!\n", fbName);
    exit(-1);
  }
  minFF = N.FB->rfb_size + N.FB->afb_size + 64 + 32;
  if (verbose)
    printf("Getting QCB of size %d...\n", qcbSize);
  generateQCB(N.FB, qcbSize); 
  if (verbose)
    printf("Determining needed info about the number field...\n");
  mpz_poly_cp(N.f, N.FB->f);
  get_g(N.T, N.FB);
  mpz_poly_discrim(D.N, N.T);
  mpz_fact_factorEasy(&D, D.N, discFact);
  printf("Monic polynomial: T="); mpz_poly_print(stdout, "", N.T);
  /* If it the discrim. didn't completely factor above, there's
     no sense trying again! */
  getIntegralBasis(&N, &D, 0); 

  printf("Obtained integral basis:\nW = \n");
  mpz_mat_print(stdout, N.W);
  printf("denominator = "); mpz_out_str(stdout, 10, N.W_d); printf("\n");

  set_prelF(&prelF, DEFAULT_MAX_FILESIZE);

  maxSize = 0;
  for (i=0; i<prelF.numFiles; i++) {
    sprintf(prelName, "%s.%d", prelF.prefix, i);
    if (stat(prelName, &fileInfo)) {
      printf("Warning: Could not stat processed file %s. Is this the first run?.\n", prelName);
      oldSize = 0;
    } else {
      oldSize = fileInfo.st_size;
      printf("Existing file %s %1.5lfMB.\n", prelName, (double)oldSize/(1024.0*1024.0));
    }
    maxSize = MAX(maxSize, oldSize);
  }

  numNewRels=0;
  if (stat(newRelName, &fileInfo)) {
    printf("Warning: Could not stat new file %s. No new relations will be processed.\n", newRelName);
    newSize = 0;
  } else {
    newSize = fileInfo.st_size;
    printf("     New file is %1.5lfMB.\n", (double)newSize/(1024.0*1024.0));
    if ((fp = fopen(newRelName, "r"))) {
      while (!(feof(fp))) {
        fgets(tmpStr, 1023, fp);
        numNewRels++;
      }
      fclose(fp);
    }
    printf("     New file appears to have %ld relations.\n", numNewRels);
  }

  /* This is pretty ad-hoc; the point is that it's a reasonably
     safe upper bound to say that each relation needs at most 150 longs.
     In fact, we could likely get away with even 125 or so, but we
     should wait and see how it looks in practice.
  */
  RL.numRels = 0;
  RL.maxDataSize = 1000 + maxSize/sizeof(long) + 1.1*numNewRels*150/prelF.numFiles;
  if (!(RL.relData = (long *)malloc(RL.maxDataSize * sizeof(long)))) {
    fprintf(stderr, "Error allocating %ldMB for all relations!\n",
            RL.maxDataSize * sizeof(long)/1048576);
    fprintf(stderr, "Try increasing DEFAULT_NUM_FILES and re-running.\n");
    fprintf(stderr, "But this will require removing old proc files and reprocessing all relations.\n");
    exit(-1);
  }
  printf("Allocated %ldMB for relations.\n", RL.maxDataSize*sizeof(long)/1048756);
  /* Again: it's a safe bet that any relation needs at least 20 longs, so: */
  RL.maxRels = (unsigned long)RL.maxDataSize/20;
  if (!(RL.relIndex = (long *)malloc(RL.maxRels * sizeof(long)))) {
    fprintf(stderr, "Error allocating %ldMB for relation pointers!\n",
            RL.maxRels * sizeof(long)/1048756);
    free(RL.relData);
    exit(-1);
  }

  totalRels = 0;
  rStart = sTime();
  for (i=0; i<prelF.numFiles; i++) {
    sprintf(prelName, "%s.%d", prelF.prefix, i);
    RL.numRels = 0;
    if (newSize) {
      readRelList(&RL, prelName);
    } else {
      /* All we really need is the number of relations: */
      FILE *tfp;
      if ((tfp = fopen(prelName, "rb"))) {
        readRawLong(&RL.numRels, tfp);
        fclose(tfp);
      }
    }
    printf("Read %ld relations from %s.\n", RL.numRels, prelName);

    if (newSize)  {
      newSize = addNewRelations4(&RL, newRelName, i, prelF.numFiles, &N);
      printf("Read %ld good relations from %s.\n", newSize, newRelName);
      bytesUsed = sizeof(long)*RL.relIndex[RL.numRels];
      printf("Used %ldMB of %ldMB allocated for this file.\n", 
              bytesUsed/1048576, sizeof(long)*RL.maxDataSize/1048576);

      /* Rp is a temporary beast, for sorting. It will be freed within this loop also.     */
      /* I don't know why I did it this way, but this whole mess should be changed anyway. */
      if (!(Rp = (rel_t *)malloc(sizeof(rel_t)*RL.numRels))) {
        fprintf(stderr, "Error allocating %ldMB for Rp.\n", sizeof(rel_t)*RL.maxRels/1048576);
        free(RL.relData); free(RL.relIndex);
        exit(-1);
      }
      prepRelFilter(Rp, &RL);
      /* This will sort on (a,b) with rev. lex, remove duplicates,
         and write them back to the proper file.
      */
      filterRelations(Rp, &prelF, i, &RL);
    }
    printf("After filtering, there are %ld unique relations in %s.%d\n", RL.numRels, prelF.prefix,i);
    totalRels += RL.numRels;
    if (Rp != NULL) {
      free(Rp);
      Rp=NULL;
    }
  }
  rStop = sTime();
  printf("Elapsed time for processing new relations: %1.4lf seconds.\n", rStop-rStart);
  free(RL.relData); free(RL.relIndex);
  msgLog("", "RelProcTime: %1.1lf", rStop-rStart);

  printf("********************************\n");
  printf("There are now a total of %ld unique relations in %d files.\n",
          totalRels, prelF.numFiles);
  initialRelations = totalRels;


  printf("Calling getCols()...\n");
  finalFF = getCols("cols", &prelF, &lpF, &N, totalRels, minFF);
  if (finalFF > 0)
    msgLog("", "rels:%ld, initialFF:%ld, finalFF:%ld", 
           initialRelations, initialFF, finalFF);
  if (finalFF < minFF) {
    printf("More columns needed (current = %ld, min = %ld)\n",
           finalFF, minFF);
    exit(0);
  }

  /* Write some header information that will eventually be needed
     in the `deps' file.
  */
  if (!(fp = fopen("depinf", "wb"))) {
    fprintf(stderr, "Error opening %s for write!\n", "depinf");
  } else {
    sprintf(str, "NUMCOLS: %8.8lx", finalFF); writeBinField(fp, str);
    sprintf(str, "COLNAME: %s.index", colName); writeBinField(fp, str);
    sprintf(str, "MAXRELS: %8.8lx", totalRels); writeBinField(fp, str);
    sprintf(str, "RELPREFIX: %s", prelF.prefix); writeBinField(fp, str);
    sprintf(str, "RELFILES: %x", prelF.numFiles); writeBinField(fp, str);
    sprintf(str, "LPFPREFIX: %s", lpF.prefix); writeBinField(fp, str);
    sprintf(str, "LPFFILES: %x", lpF.numFiles); writeBinField(fp, str);
    sprintf(str, "END_HEADER"); writeBinField(fp, str);
    fclose(fp);
  }
  printf("`depinf' written. You can now run matsolve.\n");
  msgLog("", "depinf file written. Run matsolve.");

  stopTime = sTime();
  printf("Total elapsed time: %1.2lf seconds.\n", stopTime-startTime);

  return retVal;
}  


