/**************************************************************/
/* llist.c                                                    */
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Routines for efficient operation on (large) arrays
   with a variable number of longs in each entry.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include "ggnfs.h"


#define MAXFIELDENTRIES 2048


/*************************************************************/
int ll_init(llist_t *L, long maxDataSize, long maxFields)
/*************************************************************/
{
  L->data = L->index = NULL;
  L->maxDataSize = L->maxFields = 0;
  L->numFields = 0;
  if (!(L->data = (long *)malloc(maxDataSize*sizeof(long)))) {
    fprintf(stderr, "ll_init() Memory allocation error (%ldMb).\n", 
            maxDataSize*sizeof(long)/(1024*1024));
    return -1;
  }
  if (!(L->index = (long *)malloc((maxFields+1)*sizeof(long)))) {
    fprintf(stderr, "ll_init() Memory allocation error (%ldMb).\n", 
            maxFields*sizeof(long)/(1024*1024));
    free(L->data);
    L->data=NULL;
    return -1;
  }
  L->maxDataSize = maxDataSize;
  L->maxFields = maxFields;
  L->index[0]=0;
  return 0;
}

/*************************************************************/
void ll_clear(llist_t *L)
/*************************************************************/
{
  free(L->data);
  free(L->index);
  L->data = L->index = NULL;
  L->maxDataSize = 0;
  L->maxFields = 0;
  L->numFields = 0;
}

/*************************************************************/
void ll_resize(llist_t *L, long maxDataSize)
/*************************************************************/
{ long *tmp;

  tmp = realloc(L->data, maxDataSize*sizeof(long));
  if (tmp != NULL) {
    L->data = tmp;
  } else {
    printf("ll_resize() Memory (re-)allocation error!\n");
  }
  L->maxDataSize = maxDataSize;
}

#define RESIZE 1024*1024
/*********************************************************/
int ll_appendField(llist_t *L, long *entries, long numEntries)
/*********************************************************/
{ long cSize = L->index[L->numFields];
  long *tmpPtr, i;

  if (cSize + numEntries >= L->maxDataSize) {
    tmpPtr = (long *)realloc(L->data, (L->maxDataSize + numEntries + RESIZE)*sizeof(long));
    if (tmpPtr != NULL) L->data = tmpPtr;
    else {
      fprintf(stderr, "ll_appendField() Memory (re-)allocation error!\n");
      return -1;
    }
    L->maxDataSize += numEntries + RESIZE;
  }
  if (L->numFields + 2 >= L->maxFields) {
    tmpPtr = (long *)realloc(L->index, (L->maxFields + RESIZE/4)*sizeof(long));
    if (tmpPtr != NULL) L->index = tmpPtr;
    else {
      fprintf(stderr, "ll_appendField() Memory (re-)allocation error!\n");
      return -1;
    }
    L->maxFields += RESIZE/4;
  }
  for (i=0; i<numEntries; i++)
    L->data[cSize+i] = entries[i];
  L->numFields += 1;
  L->index[L->numFields] = cSize+numEntries;
  return 0;
}
  

/*********************************************************/
int ll_verify(llist_t *L)
/*********************************************************/
/* Check that the llist_t appears to be valid.           */
/*********************************************************/
{ long i, s0, s1;

  if ((L->numFields < 0) || (L->numFields > L->maxFields)) {
    printf("ll_verify() L->numFields=%ld vs. L->maxFields=%ld.\n",
           L->numFields, L->maxFields);
    return -1;
  }

  for (i=0; i<L->numFields; i++) {
    s0 = L->index[i];
    s1 = L->index[i+1];
    if (s1 > L->maxDataSize) {
      printf("ll_verify() L->index[%ld]=%ld vs. L->maxDataSize=%ld.\n",
             i+1, s1, L->maxDataSize);
      return -1;
    }
    if (s1 < s0) {
      printf("ll_verify() L->index[%ld]=%ld vs. L->index[%ld+1]=%ld.\n",
             i, s0, i, s1);
      return -1;
    }
  }
  return 0;
}
      


/*********************************************************/
int ll_deleteFields(llist_t *L, long *fields, long numFields)
/*********************************************************/
/* Delete the specified fields from the list. The        */
/* remaining fields will be shifted down to fill in the  */
/* gaps.                                                 */
/* `fields' should be sorted and unique!!!!!!!           */
/*********************************************************/
{ long destField, c0, numMove, dataSize, shift, j, m=numFields;
  int  i;
                                                                                   
  destField=fields[0];

  /* Sanity check: */
  if ((L->numFields > L->maxFields) || (L->index[L->numFields] > L->maxDataSize)) {
    printf("ll_deleteFields(): 'L' appears corrupt:\n");
    exit(-1);
  }                                                                                   
  for (i=0; i<m; i++) {
    c0 = fields[i]+1;
    numMove = (i<(m-1)) ? (fields[i+1]-c0):(L->numFields - c0);
    /* Now, move the numMove fields beginning at c0  */
    /* to destField.                                 */
    if (numMove > 0) {
      /* First the data. */
      dataSize = L->index[c0+numMove] - L->index[c0];
      memmove(&L->data[L->index[destField]], &L->data[L->index[c0]], dataSize*sizeof(long));
      /* Now fix the index fields. */
      shift = L->index[c0] - L->index[destField];
      for (j=0; j<=numMove; j++)
        L->index[c0+j] -= shift;
      for (j=0; j<=numMove; j++)
        L->index[destField+j] = L->index[c0+j];
      destField += numMove;
    }
  }
  L->numFields -= m;
  return 0;
}

/************************************************************/
int ll_numCommonEntries(llist_t *L, long field0, long field1)
/************************************************************/
/* return the number of common entries in these two fields. */
/* The fields are assumed to have a fairly small number of  */
/* entries (i.e., say, less than 20 or so).                 */
/************************************************************/
{ long i, j, num=0;

  for (i=L->index[field0]; i<L->index[field0+1]; i++)
    for (j=L->index[field1]; j<L->index[field1+1]; j++)
      if (L->data[i]==L->data[j])
        num++;
  return num;
}
  


/*********************************************************/
int ll_catFields(llist_t *L, lpair_t *pairs, long numPairs, int mod2)
/*********************************************************/
/* For each f1=pairs[i].x, f2=pairs[i].y, concatenate the*/
/* data from field f2 onto field f1. If `mod2' is        */
/* nonzero, pairs of entries will be assumed to cancel,  */
/* so there is at most one copy of each resulting data   */
/* element in the field f1.                              */
/*   This is tricky, because some of the data may shift  */
/* in one direction while some shifts in the opposite.   */
/* But we handle this with the elegance it deserves.     */
/*    It is assumed that the list of pairs is sorted     */
/* ascending on the 'x' field, and the 'x' fields are    */
/* unique! (i.e., they should be strictly increasing).   */
/*********************************************************/
{ long     i, j, c0, c1, shift, maxShift, indexShift;
  long     newEntries[MAXFIELDENTRIES], numNewEntries, numOldEntries;
  long     sIndex, dIndex, moveSize;
  long    *tmpPtr;

  if (numPairs <= 0) return 0;

  /* Compute the distance that the fields must be shifted. */
  shift=maxShift=0;
  for (i=0; i<numPairs; i++) {
    c0 = pairs[i].x;
    c1 = pairs[i].y;
    numNewEntries=0;
    for (j=L->index[c0]; (j<L->index[c0+1]) && (numNewEntries < MAXFIELDENTRIES); j++)
      newEntries[numNewEntries++] = L->data[j];
    for (j=L->index[c1]; (j<L->index[c1+1]) && (numNewEntries < MAXFIELDENTRIES); j++)
      newEntries[numNewEntries++] = L->data[j];
    if (numNewEntries >= MAXFIELDENTRIES) {
      printf("MAXFIELDENTRIES exceeded (i=%ld, c0=%ld, c1=%ld.). Ignoring...\n",i,c0,c1);
      return -1;
    }
    if (mod2)
      numNewEntries = removeLongPairs(newEntries, numNewEntries);
    shift +=  numNewEntries - (L->index[c0+1] - L->index[c0]);
    maxShift = MAX(labs(shift), maxShift);
  }
  maxShift+=1;

  /* The idea is this: Shift everything to the right by the maxShift.
     Then, we can work by simply shifting everything left
     one block at a time and not have to worry about handling
     situations where some data is overwritten. The cost is nothing:
     each byte of data will be moved twice, instead of once. But moving
     it only once would come at a huge coding complexity cost, so this
     is probably even better.
  */
  if (L->index[L->numFields] + maxShift >= L->maxDataSize) {
    /* We need to do a reallocation: */
    tmpPtr = realloc(L->data, (L->index[L->numFields] + maxShift + 10)*sizeof(long));
    if (tmpPtr == NULL) {
      printf("ll_catFields(): memory reallocation error!\n");
      free(pairs); return -1;
    }
    L->data = tmpPtr;
    L->maxDataSize = L->index[L->numFields] + maxShift + 10;
  }
  /* Move everything right by `maxShift'. Of course, there's no need 
     to move the first block (i.e., fields left of the first field
     to be changed.)
  */
  sIndex = L->index[pairs[0].x+1];
  dIndex = sIndex + maxShift;
  moveSize = L->index[L->numFields] - sIndex;
  memmove(&L->data[dIndex], &L->data[sIndex], moveSize*sizeof(long));

  /* Now, do the field additions, moving blocks back to the left as we go: */
  indexShift = 0;
  for (i=0; i<numPairs; i++) {
    /* Right now, the fields in [0, pairs[i].x] are in proper position,
       starting at L->data[0], and they all have correct L->index values.
       The fields in [pairs[i].x + 1, ..., L->numFields) are all shifted
       in memory by `maxShift', and their corresponding L->index values
       are off by that amount as well.
    */

    c0 = pairs[i].x; c1 = pairs[i].y;
    /* Compute the new entries for column c0: */
    numNewEntries=0;
    for (j=L->index[c0]; j<L->index[c0+1]+indexShift; j++)
      newEntries[numNewEntries++] = L->data[j];
    numOldEntries = numNewEntries;
    if (c1 <=c0) {
      for (j=L->index[c1]; j<L->index[c1+1]; j++)
        newEntries[numNewEntries++] = L->data[j];
    } else {
      for (j=L->index[c1]; j<L->index[c1+1]; j++)
        newEntries[numNewEntries++] = L->data[j+maxShift];
    }
    if (mod2)
      numNewEntries = removeLongPairs(newEntries, numNewEntries);
    /* Place the new entries: */
    for (j=0; j<numNewEntries; j++)
      L->data[L->index[c0]+j] = newEntries[j];

    /* Move the next block of data and adjust it's indicies. */
    indexShift += numNewEntries - numOldEntries;
    sIndex = L->index[c0+1] + maxShift;
    dIndex = L->index[c0]+numNewEntries;
    if (i<(numPairs-1)) {
      moveSize = L->index[pairs[i+1].x+1] - sIndex + maxShift;
      memmove(&L->data[dIndex], &L->data[sIndex], moveSize*sizeof(long));
      for (j=c0+1; j<=pairs[i+1].x; j++) 
        L->index[j] += indexShift;
    } else {
      moveSize = L->index[L->numFields] - sIndex + maxShift;
      memmove(&L->data[dIndex], &L->data[sIndex], moveSize*sizeof(long));
      for (j=c0+1; j<=L->numFields; j++) 
        L->index[j] += indexShift;
    }
  }  
  return 0;
}


/************************************************************/
/* These are for functions which need access to a llist_t,  */
/* but for some reason or another, it's not convenient to   */
/* pass it as a pointer to the function.                    */
/************************************************************/
static llist_t *llistPtr=NULL;
void ll_setGlobal(llist_t *L)
{ llistPtr = L; }
/****************************************************/
int ll_cmpFieldSize(const void *A, const void *B)
/****************************************************/
/* Uses the global list set by ll_setGlobal().      */
/****************************************************/
{ long *a=(long *)A, *b=(long *)B;
  long s1, s2;

  s1 = llistPtr->index[*a+1] - llistPtr->index[*a];
  s2 = llistPtr->index[*b+1] - llistPtr->index[*b];
  if (s1 < s2) return -1;
  if (s1 > s2) return 1;
  return 0;
}

/***************************************************************/
int ll_getsortOnFieldSize(long *fields, llist_t *L)
/***************************************************************/
/* `fields' is allocated for at least L->numFields longs.      */
/* On output, fields[0], fields[1],... will be indicies to the */
/* fields of 'L', sorted in ascending order according to the   */
/* number of entries in each field.                            */
/***************************************************************/
{ long i;

  for (i=0; i<L->numFields; i++)
    fields[i]=i;
  /* This will sort the field number's based on the number of entries
     in each field of L.
  */
  ll_setGlobal(L);
  qsort(fields, L->numFields, sizeof(long), ll_cmpFieldSize);
  return 0;
}

