/**************************************************************/
/* matolve.c                                                  */
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>


#ifndef _MSC_VER 
#include <sys/time.h>
#endif
#include "ggnfs.h"

#define DEFAULT_SEED 1
#define DEFAULT_DEPNAME "deps"
#define DEFAULT_COLNAME "cols"


#define USAGE "[OPTIONS]\n"\
"-v           : verbose.\n"\
"-seed <int>  : Set the seed for the PRNG.\n"\
"--help       : Show this help and quit.\n"

#define START_MSG \
"\n"\
" __________________________________________________________ \n"\
"|        This is the matsolve program for GGNFS.           |\n"\
"| Version: %-15s                                 |\n"\
"| This program is copyright 2004, Chris Monico, and subject|\n"\
"| to the terms of the GNU General Public License version 2.|\n"\
"|__________________________________________________________|\n"


/***********************************************************/
/* This structure is for columns that are being processed. */
/* Before processing, we know only the relations that      */
/* comprise each column. After processing, we know all the */
/* nonzero rows of the matrix (i.e., the primes appearing  */
/* with odd exponent and the QCB values and sign value).   */
/* During processing, we will know some of each. That is,  */
/* we might have already replaced some relations with their*/
/* factorizations, but not yet all of them. These data are */
/* what will be stored in the file cols.np.                */
/***********************************************************/
typedef struct {
  char numRels;
  int  numPrimes;
  long Rels[MAX_RELS_IN_FF];
  long QCB[2];
  char sign;
  long rows[MAX_ROWS_IN_COL];
} column_t;




/***** Globals *****/
int verbose=0;

/******************************************************/
void writeColIndex(FILE *fp, column_t *C)
/******************************************************/
/* This is the indicies of the relations contributing */
/* to the column. We don't care about anything else.  */
/******************************************************/
{ 
  fwrite(&C->numRels, sizeof(char), 1, fp);
  if (C->numRels > 0)
    fwrite(&C->Rels, sizeof(long), C->numRels, fp);
}
/******************************************************/
void readColIndex(column_t *C, FILE *fp)
/******************************************************/
/* This is the indicies of the relations contributing */
/* to the column. We don't care about anything else.  */
/******************************************************/
{ 
  fread(&C->numRels, sizeof(char), 1, fp);
  if (C->numRels > 0)
    fread(&C->Rels, sizeof(long), C->numRels, fp);
}

/******************************************************/
void writeColLF(FILE *fp, column_t *C)
/******************************************************/
{ 
  fwrite(&C->numRels, sizeof(char), 1, fp);
  fwrite(&C->numPrimes, sizeof(int), 1, fp);
  if (C->numRels > 0)
    fwrite(&C->Rels, sizeof(long), C->numRels, fp);
  fwrite(&C->QCB, sizeof(long), 2, fp);
  fwrite(&C->rows, sizeof(long), C->numPrimes, fp);
}

/******************************************************/
int readColLF(column_t *C, FILE *fp)
/******************************************************/
{ 
  fread(&C->numRels, sizeof(char), 1, fp);
  fread(&C->numPrimes, sizeof(int), 1, fp);
  if (C->numRels > 0)
    fread(&C->Rels, sizeof(long), C->numRels, fp);
  fread(&C->QCB, sizeof(long), 2, fp);
  if (feof(fp)) return -1;
  fread(&C->rows, sizeof(long), C->numPrimes, fp);
  return 0;
}

/******************************************************/
void writeColSF(FILE *fp, column_t *C)
/******************************************************/
/* For these, all we care about is the prime-indexing */
/* fields and nothing else.                           */
/******************************************************/
{ 
  fwrite(&C->numPrimes, sizeof(int), 1, fp);
  fwrite(&C->QCB, sizeof(long), 2, fp);
  fwrite(&C->rows, sizeof(long), C->numPrimes, fp);
}

/******************************************************/
void readColSF(column_t *C, FILE *fp)
/******************************************************/
{ 
  fread(&C->numPrimes, sizeof(int), 1, fp);
  fread(&C->QCB, sizeof(long), 2, fp);
  if (C->numPrimes < MAX_ROWS_IN_COL) 
    fread(&C->rows, sizeof(long), C->numPrimes, fp);
  else {
    printf("readColSF() : MAX_ROWS_IN_COL exceeded! Cannot continue!\n");
    exit(-1);
  }
}


/*********************************************************/
long loadMat(nfs_sparse_mat_t *M, char *colName)
/*********************************************************/
{ FILE  *fp;
  long   i, j, k, fileSize, index, nR, r, *rwt, blockWt;
  struct stat fileInfo;
  column_t     C;


  if (stat(colName, &fileInfo)) {
    printf("loadMat() Could not stat %s!\n", colName);
    return -1;
  }
  fileSize = fileInfo.st_size;

  if (!(fp = fopen(colName, "rb"))) {
    fprintf(stderr, "loadMat() Unable to open %s for read!\n", colName);
    return -1;
  }
  fread(&M->numCols, sizeof(long), 1, fp);
  /* Scan the file once to find out how many dense rows there are. */
  rwt = (long *)malloc(M->numCols*sizeof(long));
  memset(rwt, 0x00, M->numCols*sizeof(long));

  nR = 0;
  for (j=0; j<M->numCols ; j++) {
    C.numPrimes = 0;
    readColSF(&C, fp);
    for (i=0; i<C.numPrimes; i++) {
      r = C.rows[i];
if (r >= M->numCols) {
 printf("Error: Column %ld has entry of %ld, versus M->numCols=%ld!\n",
         j, r, M->numCols);
 exit(-1);
}
      rwt[r] += 1;
      nR = MAX(nR, r);
    }
  }
  M->numRows = nR + 1 + 64; /* '1' b/c we count from zero. '64' for the additional 2 QCB words. */
  /* We have at least two dense blocks to start: */
  M->numDenseBlocks = 2;
  M->denseBlockIndex[0] = M->numRows - 64;
  M->denseBlockIndex[1] = M->numRows - 32;
printf("Matrix scanned: it should be %ld x %ld.\n", M->numRows, M->numCols);

  i=0;
  /* This could be made a bit slicker, but whatever. */
  while ((i<nR-32) && (M->numDenseBlocks < MAX_DENSE_BLOCKS)) {
    /* Is the block of rows [i, i+31] dense? */
    for (j=0, blockWt=0; j<32; j++) 
      blockWt += rwt[i+j]; 
    /* This condition may need to be adjusted. '1.1' will start to save memory,
       but is probably a little slower than '2'. */
    if (blockWt > 1.2*M->numCols) {
      /* Check: Is this row fairly dense? */
      if (rwt[i] > 0.005*M->numCols) {
        /* Okay - this is a dense block. */
        M->denseBlockIndex[M->numDenseBlocks] = i;
        M->numDenseBlocks += 1;
        /* rwt is reused below, to check which entries are from dense blocks. */
        for (j=0; j<32; j++)
          rwt[i+j]=1; 
        i += 32;
      } else {
        rwt[i++]=0;
      }
    } else {
      rwt[i++]=0;
    }
  }
  printf("Found %ld dense blocks. Re-reading matrix...\n", M->numDenseBlocks);
printf("The dense blocks consist of the following sets of rows:\n");
for (k=0; k<M->numDenseBlocks; k++) 
printf("[%ld, %ld]\n", M->denseBlockIndex[k], M->denseBlockIndex[k]+31);

  rewind(fp);
  fread(&M->numCols, sizeof(long), 1, fp);

  /* We could do a little better, by discounting for the QCB and sign entries. */
  M->maxDataSize = 256 + fileSize/sizeof(long);
  if (!(M->cEntry = (long *)malloc(M->maxDataSize*sizeof(long)))) {
    fprintf(stderr, "loadMat() Error allocating %ld bytes for the sparse matrix!\n",
            M->maxDataSize*sizeof(long));
    fclose(fp); return -1;
  }
  if (!(M->cIndex = (long *)malloc((M->numCols+1)*sizeof(long)))) {
    fprintf(stderr, "loadMat() Error allocating %ld bytes for the sparse matrix indicies!\n",
            (M->numCols+1)*sizeof(long));
    free(M->cEntry); fclose(fp); return -1;
  }
  for (i=0; i<M->numDenseBlocks; i++) {
    if (!(M->denseBlocks[i] = (long *)calloc((M->numCols+1),sizeof(long)))) {
      fprintf(stderr, "loadMat() Error allocating %ld bytes for the QCB entries!\n",
              (M->numCols+1)*sizeof(long));
      free(M->cIndex); free(M->cEntry); fclose(fp); return -1;
    }
  }

  M->cIndex[0] = 0; index=0;
  for (j=0; (j<M->numCols) && ((index+50) < M->maxDataSize); j++) {
    C.numPrimes = 0;
    readColSF(&C, fp);
    for (i=0; i<C.numPrimes; i++) {
      r = C.rows[i];
      /* Is this entry in a dense block? */
      if (rwt[r]==1) {
        /* Find the block (it won't be QCB, though). */
        for (k=2; k<M->numDenseBlocks; k++) {
          if ((r>=M->denseBlockIndex[k]) && (r < 32+M->denseBlockIndex[k])) {
//printf("Entry r=%ld is in block %ld (bit # %ld).\n",r,k,r-M->denseBlockIndex[k]);
            M->denseBlocks[k][j] ^= BIT(r-M->denseBlockIndex[k]);
            break;
          }
        }
      } else {
        M->cEntry[index] = r;
        index++;
      }
    }
    M->denseBlocks[0][j] = C.QCB[0];
    M->denseBlocks[1][j] = C.QCB[1]|0x80000000; /* sign bit. */

    /* We have enough room for one extra index, and we use it */
    /* to determine size, so this is ok even on the last pass */
    /* through the loop:                                      */
    M->cIndex[j+1] = index;
  }
  fclose(fp);
  free(rwt);
  return M->numRows;
}    


/****************************************************/
int main(int argC, char *args[])
/****************************************************/
{ char       colName[64], depName[64], str[1024];
  double     startTime, stopTime;
  long       *deps, origC, seed=DEFAULT_SEED;
  struct stat fileInfo;
  nfs_sparse_mat_t M;
  int        i;
  FILE      *fp, *ifp;

  strcpy(colName, DEFAULT_COLNAME);
  strcpy(depName, DEFAULT_DEPNAME);
  printf(START_MSG, GGNFS_VERSION);
  if (stat("depinf", &fileInfo)) {
    printf("Could not stat depinf! Are you trying to run %s to soon?\n", args[0]);
    return -1;
  }

  for (i=1; i<argC; i++) {
    if (strcmp(args[i], "-v")==0) {
      verbose++;
    } else if (strcmp(args[i], "-seed")==0) {
      if ((+i) < argC)
        seed = atol(args[i]);
    } else if (strcmp(args[i], "--help")==0) {
      printf("USAGE: %s %s\n", args[0], USAGE);
      exit(0);
    }
  }
  srand(seed);
  seedBlockLanczos(seed);
  startTime = sTime();
  msgLog("", "GGNFS-%s : matsolve", GGNFS_VERSION);



  printf("Loading matrix into RAM...\n");
  loadMat(&M, colName);
  printf("Matrix loaded: it is %ld x %ld.\n", M.numRows, M.numCols);
  if (M.numCols < (M.numRows + 32)) {
    printf("More columns needed (current = %ld, min = %ld)\n",
           M.numCols, M.numRows+32);
    free(M.cEntry); free(M.cIndex);
    exit(-1);
  }

  if (!(deps = (long *)malloc(M.numCols*sizeof(long)))) {
    printf("Could not allocate %ld bytes for the dependencies.\n", M.numCols*sizeof(long));
    free(M.cEntry); free(M.cIndex); return -1;
  }
  origC = M.numCols;
  getDependencies(&M, deps);


  if (!(ifp = fopen("depinf", "rb"))) {
    fprintf(stderr, "Error opening depinf for read!\n");
    exit(-1);
  }
  printf("Writing dependencies to file %s.\n", depName);
  if (!(fp = fopen(depName, "wb"))) {
    fprintf(stderr, "Error opening %s for write!\n", depName);
  } else {
    /* Get the header information from depinf. */
    readBinField(str, 1024, ifp);
    while (!(feof(ifp)) && strncmp(str, "END_HEADER",10)) {
      writeBinField(fp, str);
      readBinField(str,1024,ifp);
    }
    if (strncmp(str, "END_HEADER",10)) {
      fprintf(stderr, "Error: depinf is corrupt!\n");
      fclose(ifp); fclose(fp); exit(-1);
    }
    writeBinField(fp, str);
    fclose(ifp);
    fwrite(deps, sizeof(long), origC, fp);
    fclose(fp);
  }

  stopTime = sTime();
  printf("Total elapsed time: %1.2lf seconds.\n", stopTime-startTime);


  free(M.cEntry); free(M.cIndex); free(deps);
  return 0;
}  
