/**************************************************************/
/* spmatmul.c                                                 */
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>


#ifndef _MSC_VER 
#include <sys/time.h>
#endif
#include "ggnfs.h"


/* These are the functions that will need to be parallelized
   to do parallelized block Lanczos. In principle it's easy,
   but in practice it will be tricky. It can be done either
   with MPI or simply by TCP. In either case, the idea is as follows:
   (1) Maintain a list of worker nodes.
   (2) Partition the columns into disjoint subsets, one per node.
       (Each column corresponds to full relation, so they have
        very roughly equal weight).
   (3) To do a mult, send the vector to each available node.
   (4) Wait for all partial results to come back. Periodically
       the worker nodes should be queried. If contact is lost, reassign
       it's task to another node (note: this means also re-sending
       the corresponding subset of colmns, so this should only be done
       if we're sure the worker node died for some reason. It also means
       that we require a fairly reliable grid of machines, or we may wind
       up creating serious bandwidth issues). I think the idea is to have
       a relatively small number of nodes, perhaps 16.
   (5) If the work is distributed properly, the final result will be the
       XOR of the sub-tasks. So just keep a running XOR of current results
       as recieved. When they have all been recieved, the calculation is
       complete.
  What does this mean in practice? With the exception of an initialization
  phase where the whole matrix is sent out one piece at a time, 
  the required bandwidth is (number of nodes)*(vector size) which is 
  not too bad. So it can probably even be done over a 10Mbs network.
  There are still some wicked details to be worked out but I will get to 
  this soon. Most likely, it will be an MPI implementation, compatible
  with LAM-MPI.
*/

void multT(long *res, long *A, long *B, long n);
void multnx32(long *C_n, long *u_n, long *w, long n);

/******************************************************/
void MultB32(long *Product, long *x, void *P)
/******************************************************/
{ nfs_sparse_mat_t *M = (nfs_sparse_mat_t *)P;
  long i, k, offset;
  long Q[32], *ptr, *stop;

  /* Although there are only M->numRows rows in the */
  /* result, we are expected to zero all of them.   */
  memset(Product, 0x00, M->numCols*sizeof(long)); 

  for (k=0; k<M->numDenseBlocks; k++) {
    multT(Q, M->denseBlocks[k], x, M->numCols);
    offset = M->denseBlockIndex[k];
    for (i=0; i<32; i++) 
      Product[offset+i] = Q[i];
  }

  ptr = M->cEntry;
  for (i=0; i<M->numCols; i++) {
    stop = &M->cEntry[M->cIndex[i+1]];
    while (ptr < stop) {
      Product[*ptr++] ^= x[i];
    }
  }
}

/******************************************************/
void MultB_T32(long *Product, long *x, void *P)
/******************************************************/
{ nfs_sparse_mat_t *M = (nfs_sparse_mat_t *)P;
  long i, k;
  long *Q, Y[32], offset, *ptr, *stop;

  memset(Product, 0x00, M->numCols*sizeof(long));

  Q = (long *)malloc(M->numCols*sizeof(long));
  for (k=0; k<M->numDenseBlocks; k++) {
    offset = M->denseBlockIndex[k];
    for (i=0; i<32; i++)
      Y[i] = x[offset+i];
    multnx32(Q, M->denseBlocks[k], Y, M->numCols);
    for (i=0; i<M->numCols; i++)
      Product[i] ^= Q[i];
  }
  free(Q);

  ptr = M->cEntry;
  for (i=0; i<M->numCols; i++) {
    stop = &M->cEntry[M->cIndex[i+1]];
    while (ptr < stop) {
      Product[i] ^= x[*ptr++];
    }
  }
}


#ifdef _BROKEN
/* These next two functions are for illustration in parallelizing.
   The idea is that if the nfs_sparse_mat_t structure is locally
   set up to `act like' a full one, the code is nearly identical.
   For example, if node 1 is assigned columns [200000, 400000),
   then she should have her 'M' structure set up so that the
   columns are indexed that way (i.e., all the others will be empty
   since the server has not informed her of them - she only knows
   about columns 200000..399999, but it is important that she keeps
   them indexed properly.

   In fact, everything below this point is for setting up parallel
   Lanczos. None of it is used (or even tested) yet.

   Note: Everything below here still needs to be modified to
   look like above (i.e., the functions above are much faster).
*/


/******************************************************/
void MultB_partial32(long *Product, long *x, nfs_sparse_mat_t *M,
                   long col0, long col1)
/******************************************************/
/* Compute the part of the product Bx correponding to */
/* columns col0 <= c < c1.                            */
/******************************************************/
{ long i, j, Q1, Q2, qOffset;

  /* Although there are only M->numRows rows in the */
  /* result, we are expected to zero all of them.   */
  memset(Product, 0x00, M->numCols*sizeof(long)); 
  for (i=col0; i<col1; i++) {
    for (j=M->cIndex[i]; j<M->cIndex[i+1]; j++) {
      Product[M->cEntry[j]] ^= x[i];
    }
    /* Now, handle the QCB entries. */
    qOffset = M->numRows - 64;
    Q1 = M->qcb1[i]; Q2 = M->qcb2[i];
    for (j=0; j<32; j++) {
      if (Q1&BIT(j))
        Product[qOffset+j] ^= x[i];
    }
    qOffset += 32; /* Move to the next set of QCB rows. */
    for (j=0; j<32; j++)  {
      if (Q2&BIT(j))
        Product[qOffset+j] ^= x[i];
    }
  }
}

/******************************************************/
void MultB_T_partial32(long *Product, long *x, nfs_sparse_mat_t *M, 
                     long col0, long col1)
/******************************************************/
/* Compute the part of the product x(B^T) correponding*/
/* to columns col0 <= c < c1.                         */
/******************************************************/
{ long i, j, Q1, Q2, qOffset;

  memset(Product, 0x00, M->numCols*sizeof(long)); 
                                                                                                              
  for (i=col0; i<col1; i++) {
    for (j=M->cIndex[i]; j<M->cIndex[i+1]; j++) {
      Product[i] ^= x[M->cEntry[j]];
    }
    qOffset = M->numRows - 64;
    Q1 = M->qcb1[i]; Q2 = M->qcb2[i];
    for (j=0; j<32; j++)
      if (Q1&BIT(j))
        Product[i] ^= x[qOffset+j];
    qOffset += 32; /* Move to the next set of QCB rows. */
    for (j=0; j<32; j++)
      if (Q2&BIT(j))
        Product[i] ^= x[qOffset+j];
  }
}

/******************************************************/
void MultB64(u64 *Product, u64 *x, void *P)
/******************************************************/
{ nfs_sparse_mat_t *M = (nfs_sparse_mat_t *)P;
  long i, j, Q1, Q2, qOffset;

  /* Although there are only M->numRows rows in the */
  /* result, we are expected to zero all of them.   */
  memset(Product, 0x00, M->numCols*sizeof(u64)); 
  for (i=0; i<M->numCols; i++) {
    for (j=M->cIndex[i]; j<M->cIndex[i+1]; j++) {
      Product[M->cEntry[j]] ^= x[i];
    }
    /* Now, handle the QCB entries. */
    qOffset = M->numRows - 64;
    Q1 = M->qcb1[i]; Q2 = M->qcb2[i];
    for (j=0; j<32; j++) {
      if (Q1&BIT(j))
        Product[qOffset+j] ^= x[i];
    }
    qOffset += 32; /* Move to the next set of QCB rows. */
    for (j=0; j<32; j++)  {
      if (Q2&BIT(j))
        Product[qOffset+j] ^= x[i];
    }
  }
}
                                                                                                              
/******************************************************/
void MultB_T64(u64 *Product, u64 *x, void *P)
/******************************************************/
{ nfs_sparse_mat_t *M = (nfs_sparse_mat_t *)P;
  long i, j, Q1, Q2, qOffset;

  memset(Product, 0x00, M->numCols*sizeof(u64)); 

  for (i=0; i<M->numCols; i++) {
    for (j=M->cIndex[i]; j<M->cIndex[i+1]; j++) {
      Product[i] ^= x[M->cEntry[j]];
    }
    qOffset = M->numRows - 64;
    Q1 = M->qcb1[i]; Q2 = M->qcb2[i];
    for (j=0; j<32; j++)
      if (Q1&BIT(j))
        Product[i] ^= x[qOffset+j];
    qOffset += 32; /* Move to the next set of QCB rows. */
    for (j=0; j<32; j++)
      if (Q2&BIT(j))
        Product[i] ^= x[qOffset+j];
  }
}


/******************************************************/
void MultB_partial64(u64 *Product, u64 *x, nfs_sparse_mat_t *M,
                   long col0, long col1)
/******************************************************/
/* Compute the part of the product Bx correponding to */
/* columns col0 <= c < c1.                            */
/******************************************************/
{ long i, j, Q1, Q2, qOffset;

  /* Although there are only M->numRows rows in the */
  /* result, we are expected to zero all of them.   */
  memset(Product, 0x00, M->numCols*sizeof(u64)); 
  for (i=col0; i<col1; i++) {
    for (j=M->cIndex[i]; j<M->cIndex[i+1]; j++) {
      Product[M->cEntry[j]] ^= x[i];
    }
    /* Now, handle the QCB entries. */
    qOffset = M->numRows - 64;
    Q1 = M->qcb1[i]; Q2 = M->qcb2[i];
    for (j=0; j<32; j++) {
      if (Q1&BIT(j))
        Product[qOffset+j] ^= x[i];
    }
    qOffset += 32; /* Move to the next set of QCB rows. */
    for (j=0; j<32; j++)  {
      if (Q2&BIT(j))
        Product[qOffset+j] ^= x[i];
    }
  }
}

/******************************************************/
void MultB_T_partial64(u64 *Product, u64 *x, nfs_sparse_mat_t *M, 
                     long col0, long col1)
/******************************************************/
/* Compute the part of the product x(B^T) correponding*/
/* to columns col0 <= c < c1.                         */
/******************************************************/
{ long i, j, Q1, Q2, qOffset;

  memset(Product, 0x00, M->numCols*sizeof(u64)); 
                                                                                                              
  for (i=col0; i<col1; i++) {
    for (j=M->cIndex[i]; j<M->cIndex[i+1]; j++) {
      Product[i] ^= x[M->cEntry[j]];
    }
    qOffset = M->numRows - 64;
    Q1 = M->qcb1[i]; Q2 = M->qcb2[i];
    for (j=0; j<32; j++)
      if (Q1&BIT(j))
        Product[i] ^= x[qOffset+j];
    qOffset += 32; /* Move to the next set of QCB rows. */
    for (j=0; j<32; j++)
      if (Q2&BIT(j))
        Product[i] ^= x[qOffset+j];
  }
}

/*******************************************************/
int setupPartialMat(nfs_sparse_mat_t *M, long *cEntries, long size, 
                    long *cIndex, long *qcb1, long *qcb2, long c0, long c1,
                    long numRows, long numCols)
/*******************************************************/
/* cEntries are the column entries for columns [c0,c1) */
/* of the matrix. `size' is the number of longs in     */
/* `cEntries'. `cIndex' is a list of indicies into     */
/* `cEntries', decribing where each column's data is,  */
/* relative to any starting position. That is, it      */
/* doesn't really matter what cIndex[0] is literally:  */
/* all that matters is that the first column uses      */
/* cIndex[1] - cIndex[0] entries, and so on.           */
/* `qcbWords' are the qcb entries, 2 longs per column. */
/* numRows and numCols are the dimensions of the full  */
/* matrix - not just the subset we are working with!   */
/* This function will just do some pointer fiddling,   */
/* but use the memory pointed to by the given pointers,*/
/* so don't de-allocate it until program completion!   */
/*******************************************************/
/* Given these inputs, we will setup 'M' so that it    */
/* works with the above functions. That is, it will    */
/* behave as if it were the full matrix in an          */
/* nfs_sparse_mat_t structure, except that it's only   */
/* defined for columns in [c0, c1).                    */
/*******************************************************/
{ 
  M->numRows = numRows; M->numCols = numCols;
  M->cEntry = cEntries -cIndex[0];
  M->cIndex = cIndex;
  M->qcb1 = qcb1 - c0;
  M->qcb2 = qcb2 - c0;
}
 
#endif
